from typing import no_type_check

from guppylang_internals.decorator import hugr_op
from guppylang_internals.std._internal.compiler.tket_exts import QSYSTEM_UTILS_EXTENSION
from guppylang_internals.std._internal.util import external_op


@hugr_op(external_op("GetCurrentShot", [], ext=QSYSTEM_UTILS_EXTENSION))
@no_type_check
def get_current_shot() -> int:
    """Get the current shot number."""



from typing import Any


class GuppyModule:
    """A Guppy module that may contain function and type definitions."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        err = (
            "Explicit Guppy modules are no longer supported. Use the regular `@guppy` "
            "decorator without passing a module and use `foo.compile()` to "
            "compile Guppy functions."
        )
        raise RuntimeError(err)



from typing import Generic, no_type_check

from guppylang_internals.decorator import custom_function, extend_type
from guppylang_internals.std._internal.compiler.option import (
    OptionConstructor,
    OptionTestCompiler,
    OptionUnwrapCompiler,
    OptionUnwrapNothingCompiler,
)
from guppylang_internals.tys.builtin import option_type_def

from guppylang import guppy
from guppylang.std.lang import owned
from guppylang.std.mem import mem_swap

L = guppy.type_var("T", copyable=False, droppable=False)


@extend_type(option_type_def)
class Option(Generic[L]):  # type: ignore[misc]
    """Represents an optional value."""

    @custom_function(OptionTestCompiler(0))
    @no_type_check
    def is_nothing(self: "Option[L]") -> bool:
        """Returns `True` if the option is a `nothing` value."""

    @custom_function(OptionTestCompiler(1))
    @no_type_check
    def is_some(self: "Option[L]") -> bool:
        """Returns `True` if the option is a `some` value."""

    @custom_function(OptionUnwrapCompiler())
    @no_type_check
    def unwrap(self: "Option[L]" @ owned) -> L:
        """Returns the contained `some` value, consuming `self`.

        Panics if the option is a `nothing` value.
        """

    @custom_function(OptionUnwrapNothingCompiler())
    @no_type_check
    def unwrap_nothing(self: "Option[L]" @ owned) -> None:
        """Returns `None` if the option is a `nothing` value, consuming `self`.

        Panics if the option is a `some` value.
        """

    @guppy
    @no_type_check
    def swap(self: "Option[L]", other: "Option[L]" @ owned) -> "Option[L]":
        """Swaps the value of `self` with `other` and returns the original value."""
        mem_swap(self, other)
        return other

    @guppy
    @no_type_check
    def take(self: "Option[L]") -> "Option[L]":
        """Swaps the value of `self` with `nothing` and returns the original value."""
        return self.swap(nothing())


@custom_function(OptionConstructor(0))
@no_type_check
def nothing() -> Option[L]:
    """Constructs a `nothing` optional value."""


@custom_function(OptionConstructor(1))
@no_type_check
def some(value: L @ owned) -> Option[L]:
    """Constructs a `some` optional value."""



"""Utilities for fixed-size arrays, denoted `array[T, n]`, for element type `T` and
compile-time constant size `n`.

See `frozenarray[T, n]` for an immutable version of the `array[T, n]` type.
"""

# mypy: disable-error-code="empty-body, misc, override, valid-type, no-untyped-def"

from __future__ import annotations

import builtins
from types import GeneratorType
from typing import TYPE_CHECKING, Generic, TypeVar, no_type_check

from guppylang_internals.decorator import custom_function, extend_type
from guppylang_internals.definition.custom import CopyInoutCompiler
from guppylang_internals.std._internal.checker import ArrayCopyChecker, NewArrayChecker
from guppylang_internals.std._internal.compiler.array import (
    ArrayDiscardAllUsedCompiler,
    ArrayGetitemCompiler,
    ArrayIsBorrowedCompiler,
    ArraySetitemCompiler,
    NewArrayCompiler,
)
from guppylang_internals.std._internal.compiler.frozenarray import (
    FrozenarrayGetitemCompiler,
)
from guppylang_internals.tys.builtin import array_type_def, frozenarray_type_def

from guppylang import guppy
from guppylang.std.err import Result, err, ok
from guppylang.std.iter import SizedIter
from guppylang.std.option import Option, nothing, some

if TYPE_CHECKING:
    from guppylang.std.lang import owned


T = guppy.type_var("T")
L = guppy.type_var("L", copyable=False, droppable=False)
n = guppy.nat_var("n")

_T = TypeVar("_T")
_n = TypeVar("_n")


@extend_type(
    array_type_def,
    # Instruct the decorator to return the original class instead of the Guppy array
    # definition. This allows us to customise the runtime behaviour of arrays in
    # comptime to behave like lists.
    return_class=True,
)
class array(builtins.list[_T], Generic[_T, _n]):
    """Sequence of homogeneous values with statically known fixed length."""

    @custom_function(ArrayGetitemCompiler())
    def __getitem__(self: array[L, n], idx: int) -> L: ...

    @custom_function(ArraySetitemCompiler())
    def __setitem__(self: array[L, n], idx: int, value: L @ owned) -> None: ...

    @guppy
    @no_type_check
    def __len__(self: array[L, n]) -> int:
        return n

    @custom_function(NewArrayCompiler(), NewArrayChecker(), higher_order_value=False)
    def __new__(): ...

    # `__new__` will be overwritten below to provide actual runtime behaviour for
    # comptime. We still need to hold on to a reference to the Guppy function so
    # `@extend_type` can find it
    __new_guppy__ = __new__

    @guppy
    @no_type_check
    def __iter__(self: array[L, n] @ owned) -> SizedIter[ArrayIter[L, n], n]:
        return SizedIter(ArrayIter(self, 0))

    @custom_function(CopyInoutCompiler(), ArrayCopyChecker())
    def copy(self: array[T, n]) -> array[T, n]:
        """Copy an array instance. Will only work if T is a copyable type."""

    @custom_function(ArrayIsBorrowedCompiler())
    def is_borrowed(self: array[L, n], idx: int) -> bool:
        """Checks if an element has been taken out of the array.

        This is the case whenever a non-copyable element is borrowed, or when an element
        is manually taken out of the array via the `take` method.

        # Example

        ```
        qs = array(qubit() for _ in range(10))
        h(qs[3])
        result("a", qs.is_borrowed(3))  # False
        q = qs.take(3).unwrap()
        result("a", qs.is_borrowed(3))  # True
        qs.put(qubit(), 3).unwrap()
        result("a", qs.is_borrowed(3))  # False
        ```
        """

    @custom_function(ArrayGetitemCompiler())
    def take(self: array[L, n], idx: int) -> L:
        """Takes an element out of the array.

        While regular indexing into an array only allows borrowing of elements, `take`
        actually *extracts* the element and transfers ownership to the caller. This
        makes this operation inherently unsafe: elements may no longer be accessed after
        they are taken out. Attempting to do so will result in a runtime panic.

        The complementary `array.put` method may be used to return an element back into
        the array to make it accessible again.

        Panics if the provided index is negative or out of bounds, or if the element has
        already been taken out.

        Also see `array.try_take` for a version of this function that does not panic if
        the element has already been taken out.

        # Example

        ```
        qs = array(qubit() for _ in range(10))
        h(qs[3])
        q = qs.take(3)
        measure(q)  # We're allowed to deallocate since we own `q`
        # h(qs[3])  # Would panic since qubit 3 has been taken out
        qs.put(qubit(), 3) # Put a fresh qubit back into the array
        h(qs[3])
        ```
        """

    @guppy
    @no_type_check
    def try_take(self: array[L, n], idx: int) -> Option[L]:
        """Tries to take an element out of the array.

        While regular indexing into an array only allows borrowing of elements, `take`
        actually *extracts* the element and transfers ownership to the caller. This
        makes this operation inherently unsafe: elements may no longer be accessed after
        they are taken out. Attempting to do so will result in a runtime panic.

        The complementary `array.put` method may be used to return an element back into
        the array to make it accessible again.

        Returns the extracted element or `nothing` if the element has already been taken
        out. Panics if the provided index is negative or out of bounds.

        # Example

        ```
        qs = array(qubit() for _ in range(10))
        h(qs[3])
        q = qs.try_take(3).unwrap()
        measure(q)  # We're allowed to deallocate since we own `q`
        # h(qs[3])  # Would panic since qubit 3 has been taken out
        qs.put(qubit(), 3)  # Put a fresh qubit back into the array
        h(qs[3])
        ```
        """
        if self.is_borrowed(idx):
            return nothing()
        return some(self.take(idx))

    @custom_function(ArraySetitemCompiler(elem_first=True))
    def put(self: array[L, n], elem: L @ owned, idx: int) -> None:
        """Puts an element back into the array if it has been taken out previously.

        This is the complement of `array.take`. It may be used to fill the "hole" left
        by `array.take` with a new element.

        Panics if the provided index is negative or out of bounds, or if there is
        already an element at the given index.

        Also see `array.try_put` for a version of this function that does not panic if
        if there is already an element at the given index.

        # Example

        ```
        qs = array(qubit() for _ in range(10))
        q = qubit()
        # qs.put(q, 3)  # Would panic as there is already a qubit at index 3
        measure(qs.take(3))  # Take it out to make space for the new one
        qs.put(q, 3)
        h(qs[3])
        ```
        """

    @guppy
    @no_type_check
    def try_put(self: array[L, n], elem: L @ owned, idx: int) -> Result[None, L]:
        """Tries to put an element back into the array if it has been taken out
        previously.

        This is the complement of `array.take`. It may be used to fill the "hole" left
        by `array.take` with a new element.

        If there is already an element at the given index, then the array will not be
        mutated and the passed replacement element will be returned back in an `err`
        variant. Panics if the provided index is negative or out of bounds.

        # Example

        ```
        qs = array(qubit() for _ in range(10))
        q = qubit()
        qs.try_put(q, 3).unwrap_nothing()  # Is `nothing` since there's a qubit at idx 3
        measure(qs.take(3))  # Take it out to make space for the new one
        qs.try_put(q, 3).unwrap()
        h(qs[3])
        ```
        """
        if not self.is_borrowed(idx):
            return err(elem)
        self.put(elem, idx)
        return ok(None)

    def __new__(cls, *args: _T) -> builtins.list[_T]:  # type: ignore[no-redef]
        # Runtime array constructor that is used for comptime. We return an actual list
        # in line with the comptime unpacking logic that turns arrays into lists.
        if len(args) == 1 and isinstance(args[0], GeneratorType):
            return list(args[0])
        return [*args]


@guppy.struct
class ArrayIter(Generic[L, n]):
    """Iterator over arrays."""

    xs: array[L, n]
    i: int

    @guppy
    @no_type_check
    def __next__(
        self: ArrayIter[L, n] @ owned,
    ) -> Option[tuple[L, ArrayIter[L, n]]]:
        if self.i < int(n):
            elem = self.xs.take(self.i)
            return some((elem, ArrayIter(self.xs, self.i + 1)))
        _array_discard_all_used(self.xs)
        return nothing()


@custom_function(ArrayDiscardAllUsedCompiler())
def _array_discard_all_used(xs: array[L, n] @ owned) -> None: ...


@extend_type(frozenarray_type_def)
class frozenarray(Generic[T, n]):
    """An immutable array of fixed static size."""

    @custom_function(FrozenarrayGetitemCompiler())
    def __getitem__(self: frozenarray[T, n], item: int) -> T: ...  # type: ignore[type-arg]

    @guppy
    @no_type_check
    def __len__(self: frozenarray[T, n]) -> int:
        return n

    @guppy
    @no_type_check
    def __iter__(self: frozenarray[T, n]) -> SizedIter[FrozenarrayIter[T, n], n]:
        return SizedIter(FrozenarrayIter(self, 0))

    @guppy
    @no_type_check
    def mutable_copy(self: frozenarray[T, n]) -> array[T, n]:
        """Creates a mutable copy of this array."""
        return array(x for x in self)


@guppy.struct
class FrozenarrayIter(Generic[T, n]):
    """Iterator for frozenarrays."""

    xs: frozenarray[T, n]  # type: ignore[type-arg]
    i: int

    @guppy
    @no_type_check
    def __next__(
        self: FrozenarrayIter[T, n],
    ) -> Option[tuple[T, FrozenarrayIter[T, n]]]:
        if self.i < int(n):
            return some((self.xs[self.i], FrozenarrayIter(self.xs, self.i + 1)))
        return nothing()



"""
Debug querying the internal state of the emulator.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Protocol, TypeVar

import numpy as np
import numpy.typing as npt
from selene_quest_plugin.state import SeleneQuestState, TracedState
from typing_extensions import Self

__all__ = [
    "NotSingleStateError",
    "PartialState",
    "PartialVector",
    "StateVector",
    "TracedState",
]

#: Type encoding a single state. e.g. StateVector for a state vector.
S = TypeVar("S")
#: Type encoding a state vector, type alias for numpy array of complex numbers.
StateVector = npt.NDArray[np.complexfloating]


@dataclass(frozen=True)
class NotSingleStateError(Exception):
    """Raised when a Selene state is not a single state, i.e. it has unspecified qubits
    that have been traced out."""

    total_qubits: int
    n_specified_qubits: int

    def __str__(self) -> str:
        return (
            f"Selene state is not a single state: "
            f"total qubits {self.total_qubits} != "
            f"specified qubits {self.n_specified_qubits}."
        )


class PartialState(Protocol[S]):
    """Protocol for an emulator state type.
    Different simulation backends may have different state representations."""

    @property
    def total_qubits(self) -> int:
        """Total number of qubits in the state."""
        ...

    @property
    def specified_qubits(self) -> list[int]:
        """List of specified qubits in the state."""
        ...

    def state_distribution(self, zero_threshold: float = 1e-12) -> list[TracedState[S]]:
        """Distribution of states after tracing out unspecified qubits.
        In general this results in a distribution of states, each with a probability.

        Args:
            zero_threshold: Threshold for considering a state amplitude to be zero.
                Defaults to 1e-12.
        Returns:
            List of traced states, each with a state vector and a probability."""
        ...

    def as_single_state(self, zero_threshold: float = 1e-12) -> S:
        """If the state is a single pure state return it.

        Args:
            zero_threshold: Threshold for considering a state amplitude to be zero.
                Defaults to 1e-12.
        Raises:
            NotSingleStateError: If the state is not a pure state.
        Returns:
            The state as a single state type.
        """
        all_states = self.state_distribution(zero_threshold)
        if len(all_states) != 1:
            raise NotSingleStateError(
                total_qubits=self.total_qubits,
                n_specified_qubits=len(self.specified_qubits),
            )
        return all_states[0].state


class PartialVector(PartialState[StateVector]):
    """Partial state vector for simulator backends with statevector representation.
    Partial in the sense that some qubits may be traced out, and the state is
    represented as a distribution of state vectors.
    """

    _inner: SeleneQuestState

    def __init__(
        self, base_state: StateVector, total_qubits: int, specified_qubits: list[int]
    ) -> None:
        """Initialize a PartialVector from a base state vector, total qubits, and
        specified qubits.

        Args:
            base_state: The state vector over all qubits in the system.
            total_qubits: Total number of qubits in the system
            specified_qubits: List of specified qubits in the state. Those not in this
            list are considered traced out.
        """
        if len(base_state) != (1 << total_qubits):
            raise ValueError(
                f"Base state vector length ({len(base_state)}) does not match "
                f"2 ** total_qubits ({1 << total_qubits})."
            )
        self._inner = SeleneQuestState(
            state=base_state,
            total_qubits=total_qubits,
            specified_qubits=specified_qubits,
        )

    @property
    def total_qubits(self) -> int:
        """Total number of qubits in the state."""
        return self._inner.total_qubits

    @property
    def specified_qubits(self) -> list[int]:
        """List of specified qubits in the state. Those not in this
        list are considered traced out."""
        return self._inner.specified_qubits

    def state_distribution(
        self, zero_threshold: float = 1e-12
    ) -> list[TracedState[StateVector]]:
        out = self._inner.get_state_vector_distribution(zero_threshold=zero_threshold)

        # force float to remove numpy types
        return [TracedState(float(st.probability), st.state) for st in out]

    @classmethod
    def _from_inner(cls, inner: SeleneQuestState) -> Self:
        """Create a PartialVector from an inner SeleneQuestState."""
        obj = cls.__new__(cls)
        obj._inner = inner
        return obj

    def __str__(self) -> str:
        #  don't print base state as it may be large
        return (
            f"PartialVector(total_qubits={self.total_qubits}, "
            f"specified_qubits={self.specified_qubits})"
        )



"""Union of two types.

Type `Either[L, R]` represents a value of either typr `L` ("left") or `R` ("right").
"""

from typing import Generic, no_type_check

from guppylang_internals.decorator import custom_function, custom_type
from guppylang_internals.std._internal.compiler.either import (
    EitherConstructor,
    EitherTestCompiler,
    EitherToOptionCompiler,
    EitherUnwrapCompiler,
    either_to_hugr,
)
from guppylang_internals.tys.param import TypeParam

from guppylang import guppy
from guppylang.std.lang import owned
from guppylang.std.option import Option

L = guppy.type_var("L", copyable=False, droppable=False)
R = guppy.type_var("R", copyable=False, droppable=False)
Droppable = guppy.type_var("Droppable", copyable=False, droppable=True)

_params = [TypeParam(0, "L", False, False), TypeParam(1, "L", False, False)]


@custom_type(either_to_hugr, params=_params)
class Either(Generic[L, R]):  # type: ignore[misc]
    """Represents a union of either a `left` or a `right` value."""

    @custom_function(EitherTestCompiler(0))
    @no_type_check
    def is_left(self: "Either[L, R]") -> bool:
        """Returns `True` for a `left` value."""

    @custom_function(EitherTestCompiler(1))
    @no_type_check
    def is_right(self: "Either[L, R]") -> bool:
        """Returns `True` for a `right` value."""

    @custom_function(EitherToOptionCompiler(0))
    @no_type_check
    def try_into_left(self: "Either[L, Droppable]" @ owned) -> Option[L]:
        """Returns the wrapped value if `self` is a `left` value, or `nothing`
        otherwise.

        This operation is only allowed if the `right` variant wraps a droppable type.
        """

    @custom_function(EitherToOptionCompiler(1))
    @no_type_check
    def try_into_right(self: "Either[Droppable, R]" @ owned) -> Option[R]:
        """Returns the wrapped value if `self` is a `right` value, or `nothing`
        otherwise.

        This operation is only allowed if the `left` variant wraps a droppable type.
        """

    @custom_function(EitherUnwrapCompiler(0))
    @no_type_check
    def unwrap_left(self: "Either[L, R]" @ owned) -> L:
        """Returns the contained `left` value, consuming `self`.

        Panics if `self` is a `right` value.
        """

    @custom_function(EitherUnwrapCompiler(1))
    @no_type_check
    def unwrap_right(self: "Either[L, R]" @ owned) -> R:
        """Returns the contained `right` value, consuming `self`.

        Panics if `self` is a `left` value.
        """


@custom_function(EitherConstructor(0))
@no_type_check
def left(val: L @ owned) -> Either[L, R]:
    """Constructs a `left` either value."""


@custom_function(EitherConstructor(1))
@no_type_check
def right(val: R @ owned) -> Either[L, R]:
    """Constructs a `right` either value."""



"""The string type."""

# mypy: disable-error-code="empty-body, misc, override, valid-type, no-untyped-def"

from __future__ import annotations

from guppylang_internals.decorator import custom_function, extend_type
from guppylang_internals.std._internal.checker import UnsupportedChecker
from guppylang_internals.tys.builtin import string_type_def


@extend_type(string_type_def)
class str:
    """A string, i.e. immutable sequences of Unicode code points."""

    @custom_function(checker=UnsupportedChecker(), higher_order_value=False)
    def __new__(x): ...



"""Utilities for iteration over collections of values."""

# mypy: disable-error-code="empty-body, misc, override, valid-type, no-untyped-def"

from __future__ import annotations

from typing import TYPE_CHECKING, Any, no_type_check

from guppylang_internals.decorator import custom_function, extend_type
from guppylang_internals.definition.custom import NoopCompiler
from guppylang_internals.tys.builtin import sized_iter_type_def

from guppylang import guppy
from guppylang.std.option import Option, nothing, some

if TYPE_CHECKING:
    from guppylang.std.lang import comptime, owned
    from guppylang.std.num import nat

L = guppy.type_var("L", copyable=False, droppable=False)
n = guppy.nat_var("n")


@extend_type(sized_iter_type_def)
class SizedIter:
    """A wrapper around an iterator type `L` promising that the iterator will yield
    exactly `n` values.

    Annotating an iterator with an incorrect size is undefined behaviour.
    """

    def __class_getitem__(cls, item: Any) -> type:
        # Dummy implementation to allow subscripting of the `SizedIter` type in
        # positions that are evaluated by the Python interpreter
        return cls

    @custom_function(NoopCompiler())
    def __new__(iterator: L @ owned) -> SizedIter[L, n]:  # type: ignore[type-arg]
        """Casts an iterator into a `SizedIter`."""

    @custom_function(NoopCompiler())
    def unwrap_iter(self: SizedIter[L, n] @ owned) -> L:
        """Extracts the actual iterator."""

    @custom_function(NoopCompiler())
    def __iter__(self: SizedIter[L, n] @ owned) -> SizedIter[L, n]:  # type: ignore[type-arg]
        """Dummy implementation making sized iterators iterable themselves."""


@guppy.struct
class Range:
    next: int
    stop: int
    step: int

    @guppy
    def __iter__(self: Range) -> Range:
        return self

    @guppy
    @no_type_check
    def __next__(self: Range) -> Option[tuple[int, Range]]:
        end = (self.next >= self.stop) if self.step >= 0 else (self.next <= self.stop)
        if end:
            return nothing()
        return some((self.next, Range(self.next + self.step, self.stop, self.step)))


@guppy
@no_type_check
def _range1(stop: int) -> Range:
    return Range(0, stop, 1)


@guppy
@no_type_check
def _range2(start: int, stop: int) -> Range:
    return Range(start, stop, 1)


@guppy
@no_type_check
def _range3(start: int, stop: int, step: int) -> Range:
    return Range(start, stop, step)


@guppy
@no_type_check
def _range_comptime(stop: nat @ comptime) -> "SizedIter[Range, stop]":  # noqa: F821 UP037
    return SizedIter(Range(0, stop, 1))


@guppy.overload(_range_comptime, _range1, _range2, _range3)
def range(start: int, stop: int = 0, step: int = 1) -> Range:
    """An iterator that yields a sequence of integers.

    Behaves like the builtin Python `range` function. Concretely, the ``i``th yielded
    number is ``start + i * step``. Numbers are yielded as long as they are

    * ``< stop`` in the case where ``step >= 0``, or
    * ``> stop`` otherwise.

    ``start`` defaults to ``0`` and ``step`` defaults to ``1``. If the provided ``stop``
    value is comptime known, then the returned iterator will have a static size
    annotation and may for example be used inside array comprehensions.
    """



from __future__ import annotations

from typing import TYPE_CHECKING, Generic, no_type_check

from guppylang.decorator import guppy
from guppylang.std.array import array
from guppylang.std.option import Option, nothing, some
from guppylang.std.platform import panic

if TYPE_CHECKING:
    from guppylang.std.lang import owned

T = guppy.type_var("T", copyable=False, droppable=False)
TCopyable = guppy.type_var("TCopyable", copyable=True, droppable=False)
MAX_SIZE = guppy.nat_var("MAX_SIZE")


@guppy.struct
class Stack(Generic[T, MAX_SIZE]):  # type: ignore[misc]
    """A last-in-first-out (LIFO) growable collection of values.

    To ensure static allocation, the maximum stack size must be specified in advance and
    is tracked in the type. For example, the `Stack[int, 10]` is a stack that can hold
    at most 10 integers.

    Use `empty_stack` to construct a new stack.
    """

    #: Underlying buffer holding the stack elements.
    #:
    #: INVARIANT: All array elements up to and including index `self.end - 1` are
    #: `option.some` variants and all further ones are `option.nothing`.
    buf: array[Option[T], MAX_SIZE]  # type: ignore[valid-type, type-arg]

    #: Index of the next free index in `self.buf`.
    end: int

    @guppy
    @no_type_check
    def __len__(self: Stack[T, MAX_SIZE]) -> int:
        """Returns the number of elements currently stored in the stack."""
        return self.end

    @guppy
    @no_type_check
    def __iter__(self: Stack[T, MAX_SIZE] @ owned) -> Stack[T, MAX_SIZE]:
        """Returns an iterator over the elements in the stack from top to bottom."""
        return self

    @guppy
    @no_type_check
    def __next__(
        self: Stack[T, MAX_SIZE] @ owned,
    ) -> Option[tuple[T, Stack[T, MAX_SIZE]]]:
        if len(self) == 0:
            self.discard_empty()
            return nothing()
        val, new_stack = self.pop()
        return some((val, new_stack))

    @guppy
    @no_type_check
    def push(self: Stack[T, MAX_SIZE] @ owned, elem: T @ owned) -> Stack[T, MAX_SIZE]:
        """Adds an element to the top of the stack.

        Panics if the stack has already reached its maximum size.
        """
        if self.end >= MAX_SIZE:
            panic("Stack.push: max size reached")
        self.buf[self.end].swap(some(elem)).unwrap_nothing()
        return Stack(self.buf, self.end + 1)

    @guppy
    @no_type_check
    def pop(self: Stack[T, MAX_SIZE] @ owned) -> tuple[T, Stack[T, MAX_SIZE]]:
        """
        Removes the top element from the stack and returns it.

        Panics if the stack is empty.
        """
        if self.end <= 0:
            panic("Stack.pop: stack is empty")
        elem = self.buf[self.end - 1].take().unwrap()
        return elem, Stack(self.buf, self.end - 1)

    @guppy
    @no_type_check
    def peek(
        self: Stack[TCopyable, MAX_SIZE] @ owned,
    ) -> tuple[TCopyable, Stack[TCopyable, MAX_SIZE]]:
        """Returns a copy of the top element of the stack without removing it.

        Panics if the stack is empty.

        Note that this operation is only allowed if the stack elements are copyable.
        """
        if self.end <= 0:
            panic("Stack.peek: stack is empty")
        elem = self.buf[self.end - 1].unwrap()
        return elem, Stack(self.buf, self.end)

    @guppy
    @no_type_check
    def discard_empty(self: Stack[T, MAX_SIZE] @ owned) -> None:
        """Discards a stack of potentially non-droppable elements assuming that the
        stack is empty.

        Panics if the stack is not empty.
        """
        if self.end > 0:
            panic("Stack.discard_empty: stack is not empty")
        for elem in self.buf:
            elem.unwrap_nothing()


@guppy
@no_type_check
def empty_stack() -> Stack[T, MAX_SIZE]:
    """Constructs a new empty stack."""
    buf = array(nothing[T]() for _ in range(MAX_SIZE))
    return Stack(buf, 0)


# Deprecated reexport
from guppylang.std.collections.priority_queue import (  # noqa: F401 E402
    PriorityQueue,
    empty_priority_queue,
)



"""EXPERIMENTAL: The `list` type"""

# ruff: noqa: E501
# mypy: disable-error-code="empty-body, misc, override, valid-type, no-untyped-def, type-arg"

from __future__ import annotations

from typing import TYPE_CHECKING, Generic

from guppylang_internals.decorator import custom_function, extend_type, hugr_op
from guppylang_internals.definition.custom import NoopCompiler
from guppylang_internals.std._internal.checker import UnsupportedChecker
from guppylang_internals.std._internal.compiler.list import (
    ListGetitemCompiler,
    ListLengthCompiler,
    ListPopCompiler,
    ListPushCompiler,
    ListSetitemCompiler,
)
from guppylang_internals.std._internal.util import unsupported_op
from guppylang_internals.tys.builtin import list_type_def

from guppylang import guppy
from guppylang.std.option import Option  # noqa: TCH001

if TYPE_CHECKING:
    from guppylang.std.lang import owned


T = guppy.type_var("T")
L = guppy.type_var("L", copyable=False, droppable=False)


@extend_type(list_type_def)
class list(Generic[T]):
    """Mutable sequence items with homogeneous types."""

    @custom_function(ListGetitemCompiler())
    def __getitem__(self: list[L], idx: int) -> L: ...

    @custom_function(ListSetitemCompiler())
    def __setitem__(self: list[L], idx: int, value: L @ owned) -> None: ...

    @custom_function(ListLengthCompiler())
    def __len__(self: list[L]) -> int: ...

    @custom_function(checker=UnsupportedChecker(), higher_order_value=False)
    def __new__(x): ...

    @custom_function(NoopCompiler())  # TODO: define via Guppy source instead
    def __iter__(self: list[L] @ owned) -> list[L]: ...

    @hugr_op(unsupported_op("pop"))
    def __next__(self: list[L] @ owned) -> Option[tuple[L, list[L]]]: ...  # type: ignore[type-arg]

    @custom_function(ListPushCompiler())
    def append(self: list[L], item: L @ owned) -> None: ...

    @custom_function(ListPopCompiler())
    def pop(self: list[L]) -> L: ...



from guppylang.std.collections.priority_queue import PriorityQueue, empty_priority_queue
from guppylang.std.collections.stack import Stack, empty_stack

__all__ = ["empty_priority_queue", "empty_stack", "PriorityQueue", "Stack"]



"""Utilities for compile-time reflection."""

# mypy: disable-error-code="empty-body, misc, override, valid-type, no-untyped-def"

from guppylang_internals.decorator import custom_function
from guppylang_internals.std._internal.checker import CallableChecker


@custom_function(checker=CallableChecker(), higher_order_value=False)
def callable(x): ...



from hugr.qsystem.result import QsysShot

from .result import EmulatorResult


class EmulatorError(Exception):
    completed_shots: EmulatorResult
    failing_shot: QsysShot
    underlying_exception: Exception | None

    def __init__(
        self,
        completed_shots: EmulatorResult,
        failing_shot: QsysShot,
        underlying_exception: Exception | None = None,
    ):
        super().__init__(str(underlying_exception))
        self.completed_shots = completed_shots
        self.failing_shot = failing_shot
        self.underlying_exception = underlying_exception

    @property
    def failed_shot_index(self) -> int:
        """The index of the shot that failed."""
        return len(self.completed_shots.results)


class EmulatorBuildError(Exception):
    underlying_exception: Exception | None

    def __init__(self, underlying_exception: Exception | None = None):
        super().__init__(
            "Building the emulator failed with the following exception: "
            + str(underlying_exception)
        )
        self.underlying_exception = underlying_exception



"""Guppy standard module for functional qsystem native operations. For the mathematical
definitions of these gates, see the guppylang.std.qsystem documentation.

These gates are the same as those in std.qsystem but use functional syntax.
"""

from typing import no_type_check

from guppylang.decorator import guppy
from guppylang.std import qsystem
from guppylang.std.angles import angle
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


@guppy
@no_type_check
def phased_x(q: qubit @ owned, angle1: angle, angle2: angle) -> qubit:
    """Functional PhasedX gate command."""
    qsystem.phased_x(q, angle1, angle2)
    return q


@guppy
@no_type_check
def zz_phase(q1: qubit @ owned, q2: qubit @ owned, angle: angle) -> tuple[qubit, qubit]:
    """Functional ZZPhase gate command."""
    qsystem.zz_phase(q1, q2, angle)
    return q1, q2


@guppy
@no_type_check
def measure_and_reset(q: qubit @ owned) -> tuple[qubit, bool]:
    """Functional measure_and_reset command."""
    b = qsystem.measure_and_reset(q)
    return q, b


@guppy
@no_type_check
def zz_max(q1: qubit @ owned, q2: qubit @ owned) -> tuple[qubit, qubit]:
    """Functional ZZMax gate command."""
    qsystem.zz_max(q1, q2)
    return q1, q2


@guppy
@no_type_check
def rz(q: qubit @ owned, angle: angle) -> qubit:
    """Functional Rz gate command."""
    qsystem.rz(q, angle)
    return q


@guppy
@no_type_check
def measure(q: qubit @ owned) -> bool:
    """Functional destructive measurement command."""
    result = qsystem.measure(q)
    return result


@guppy
@no_type_check
def qfree(q: qubit @ owned) -> None:
    """Functional qfree command."""
    qsystem.qfree(q)


@guppy
@no_type_check
def reset(q: qubit @ owned) -> qubit:
    """Functional Reset command."""
    qsystem.reset(q)
    return q



# mypy: disable-error-code="empty-body, no-untyped-def"
"""Python builtins that are not supported yet"""

from guppylang_internals.decorator import custom_function
from guppylang_internals.std._internal.checker import UnsupportedChecker


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def aiter(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def all(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def anext(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def any(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def bin(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def breakpoint(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def bytearray(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def bytes(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def chr(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def classmethod(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def compile(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def complex(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def delattr(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def dict(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def dir(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def enumerate(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def eval(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def exec(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def filter(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def format(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def frozenset(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def getattr(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def globals(): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def hasattr(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def hash(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def help(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def hex(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def id(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def input(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def isinstance(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def issubclass(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def iter(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def locals(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def map(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def max(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def memoryview(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def min(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def next(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def object(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def oct(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def open(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def ord(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def print(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def property(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def repr(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def reversed(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def set(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def setattr(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def slice(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def sorted(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def staticmethod(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def sum(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def super(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def type(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def vars(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def zip(x): ...


@custom_function(checker=UnsupportedChecker(), higher_order_value=False)
def __import__(x): ...



"""
Configuring and executing emulator instances for guppy programs.
"""

from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass, field, replace
from typing import TYPE_CHECKING, Any, cast

from hugr.qsystem.result import QsysShot
from selene_sim.backends.bundled_error_models import IdealErrorModel
from selene_sim.backends.bundled_runtimes import SimpleRuntime
from selene_sim.backends.bundled_simulators import Coinflip, Quest, Stim
from selene_sim.event_hooks import EventHook, NoEventHook
from tqdm import tqdm
from typing_extensions import Self

from .exceptions import EmulatorError
from .result import EmulatorResult

if TYPE_CHECKING:
    import datetime
    from collections.abc import Iterator
    from pathlib import Path

    from hugr.qsystem.result import TaggedResult
    from selene_core.error_model import ErrorModel
    from selene_core.runtime import Runtime
    from selene_core.simulator import Simulator
    from selene_sim.instance import SeleneInstance


@dataclass(frozen=True)
class _Options:
    _simulator: Simulator = field(default_factory=Quest)
    _runtime: Runtime = field(default_factory=SimpleRuntime)
    _error_model: ErrorModel = field(default_factory=IdealErrorModel)
    _shots: int = 1
    _shot_increment: int = 1
    _shot_offset: int = 0
    _seed: int | None = None
    _verbose: bool = False
    _timeout: datetime.timedelta | None = None
    _n_processes: int = 1
    _event_hook: EventHook = field(default_factory=NoEventHook)
    # unstable:
    _results_logfile: Path | None = None
    _display_progress_bar: bool = False


@dataclass(frozen=True)
class EmulatorInstance:
    """An emulator instance for running a compiled program.


    Returned by :py:class:`GuppyFunctionDefinition.emulator`.
    Contains configuration options for the emulator instance, such as the number of
    qubits, the number of shots, the simulator backend, and more.
    """

    _instance: SeleneInstance
    _n_qubits: int
    _options: _Options = field(default_factory=_Options)

    def _with_option(self, **kwargs: Any) -> Self:
        """Helper method to simplify setting options."""
        return replace(self, _options=replace(self._options, **kwargs))

    @property
    def n_qubits(self) -> int:
        """Number of qubits available in the emulator instance."""
        return self._n_qubits

    @property
    def shots(self) -> int:
        """Number of shots to run for each execution."""
        return self._options._shots

    @property
    def simulator(self) -> Simulator:
        """Simulation backend used for running the emulator instance."""
        return self._options._simulator

    @property
    def runtime(self) -> Runtime:
        """Runtime used for executing the emulator instance."""
        return self._options._runtime

    @property
    def error_model(self) -> ErrorModel:
        """Device error model used for the emulator instance."""
        return self._options._error_model

    @property
    def verbose(self) -> bool:
        """Whether to print verbose output during the emulator execution."""
        return self._options._verbose

    @property
    def timeout(self) -> datetime.timedelta | None:
        """Timeout for the emulator execution, if any."""
        return self._options._timeout

    @property
    def seed(self) -> int | None:
        """Random seed for the emulator instance, if any."""
        return self._options._seed

    @property
    def shot_offset(self) -> int:
        """Offset for the shot numbers, shot counts will begin at this offset.
        Defaults to 0.

        This is useful for running multiple emulator instances in parallel"""
        return self._options._shot_offset

    @property
    def shot_increment(self) -> int:
        """Value to increment shot numbers by for each repeated run.
        Defaults to 1."""
        return self._options._shot_increment

    @property
    def n_processes(self) -> int:
        """Number of processes to parallelise the emulator execution across.
        Defaults to 1, meaning no parallelisation."""
        return self._options._n_processes

    def with_n_qubits(self, value: int) -> Self:
        """Set the number of qubits available in the emulator instance."""
        return replace(self, _n_qubits=value)

    def with_shots(self, value: int) -> Self:
        """Set the number of shots to run for each execution.
        Defaults to 1."""
        return self._with_option(_shots=value)

    def with_simulator(self, value: Simulator) -> Self:
        """Set the simulation backend used for running the emulator instance.
        Defaults to statevector simulation."""
        return self._with_option(_simulator=value)

    def with_runtime(self, value: Runtime) -> Self:
        """Set the runtime used for executing the emulator instance.
        Defaults to SimpleRuntime."""
        return self._with_option(_runtime=value)

    def with_error_model(self, value: ErrorModel) -> Self:
        """Set the device error model used for the emulator instance.
        Defaults to IdealErrorModel (no errors)."""
        return self._with_option(_error_model=value)

    def with_event_hook(self, value: EventHook) -> Self:
        """Set the event hook used for the emulator instance.
        Defaults to NoEventHook."""
        return self._with_option(_event_hook=value)

    def with_verbose(self, value: bool) -> Self:
        """Set whether to print verbose output during the emulator execution.
        Defaults to False."""
        return self._with_option(_verbose=value)

    def with_progress_bar(self, value: bool = True) -> Self:
        """Set whether to display a progress bar during the emulator execution.
        Defaults to False."""
        return self._with_option(_display_progress_bar=value)

    def with_timeout(self, value: datetime.timedelta | None) -> Self:
        """Set the timeout for the emulator execution.
        Defaults to None (no timeout)."""
        return self._with_option(_timeout=value)

    def with_seed(self, value: int | None) -> Self:
        """Set the random seed for the emulator instance.
        Defaults to None."""
        new_options = replace(self._options, _seed=value)
        # TODO flaky stateful, remove when selene simplifies
        new_options._simulator.random_seed = value
        out = replace(self, _options=new_options)
        return out

    def with_shot_offset(self, value: int) -> Self:
        """Set the offset for the shot numbers, shot counts will begin at this offset.
        Defaults to 0.

        This is useful for running multiple emulator instances in parallel."""
        return self._with_option(_shot_offset=value)

    def with_shot_increment(self, value: int) -> Self:
        """Set the value to increment shot numbers by for each repeated run.
        Defaults to 1."""
        return self._with_option(_shot_increment=value)

    def with_n_processes(self, value: int) -> Self:
        """Set the number of processes to parallelise the emulator execution across.
        Defaults to 1, meaning no parallelisation."""
        return self._with_option(_n_processes=value)

    def statevector_sim(self) -> Self:
        """Set the simulation backend to the default statevector simulator."""
        return self.with_simulator(Quest())

    def coinflip_sim(self) -> Self:
        """Set the simulation backend to the coinflip simulator.
        This performs no quantum simulation, and flips a coin for each measurement."""
        return self.with_simulator(Coinflip())

    def stabilizer_sim(self) -> Self:
        """Set the simulation backend to the stabilizer simulator.
        This only works for clifford circuits but is very fast."""
        return self.with_simulator(Stim())

    def run(self) -> EmulatorResult:
        """Run the emulator instance and return the results.
        By default runs one shot, this can be configured with `with_shots()`."""
        result_stream = self._run_instance()

        all_results: list[QsysShot] = []
        for shot in self._iterate_shots(result_stream):
            shot_results = QsysShot()
            try:
                for tag, value in shot:
                    shot_results.append(tag, value)
            except Exception as e:  # noqa: BLE001
                # In this case, casting a wide net on exceptions is
                # suitable.
                raise EmulatorError(
                    completed_shots=EmulatorResult(all_results),
                    failing_shot=shot_results,
                    underlying_exception=e,
                ) from None
            all_results.append(shot_results)
        return EmulatorResult(all_results)

    def _run_instance(self) -> Iterator[Iterator[TaggedResult]]:
        """Run the Selene instance with the given simulator lazily."""
        return self._instance.run_shots(
            simulator=self.simulator,
            runtime=self.runtime,
            n_qubits=self.n_qubits,
            n_shots=self.shots,
            event_hook=self._options._event_hook,
            error_model=self.error_model,
            verbose=self.verbose,
            timeout=self.timeout,
            results_logfile=self._options._results_logfile,
            random_seed=self.seed,
            shot_offset=self.shot_offset,
            shot_increment=self.shot_increment,
            n_processes=self.n_processes,
        )

    def _iterate_shots(
        self, result_stream: Iterator[Iterator[TaggedResult]]
    ) -> Iterator[Iterator[TaggedResult]]:
        """Iterate over the shots in the result stream, optionally displaying a progress
        bar."""
        if self._options._display_progress_bar:
            return cast(
                "Iterator[Iterator[TaggedResult]]",
                tqdm(result_stream, total=self.shots, desc="Emulating shots"),
            )
        else:
            return result_stream



"""Provides Python objects for builtin language keywords."""

from typing import Any, Protocol, TypeVar

T = TypeVar("T")


class _Comptime:
    """Dummy class to support `@comptime` annotations and `comptime(...)` expressions"""

    def __call__(self, v: T) -> T:
        return v

    def __rmatmul__(self, other: Any) -> Any:
        # This method is to make the Python interpreter happy with @comptime at runtime
        return other


#: Function to tag compile-time evaluated Python expressions in a Guppy context.
#:
#: This function acts like the identity when execute in a Python context.
comptime = _Comptime()


#: Alias for `comptime` expressions
py = comptime


class _Owned:
    """Dummy class to support `@owned` annotations."""

    def __rmatmul__(self, other: Any) -> Any:
        return other


owned = _Owned()


class Copy(Protocol):
    """Bound to mark generic type parameters as being implicitly copyable."""


class Drop(Protocol):
    """Bound to mark generic type parameters as being implicitly droppable."""



"""
Emulation results and post-processing.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from hugr.qsystem.result import QsysResult, QsysShot, TaggedResult
from selene_sim.backends.bundled_simulators import Quest

from .state import PartialVector

if TYPE_CHECKING:
    from collections import Counter

    from hugr.qsystem.result import DataValue
    from pytket.backends.backendresult import BackendResult
    from selene_quest_plugin.state import SeleneQuestState

# Re-exports QsysResult, QsysShot, and TaggedResult - breaking changes to those upstream
# classes should be treated as breaking changes to this module.
__all__ = [
    "EmulatorResult",
    "QsysShot",
    "TaggedResult",
]


class EmulatorResult(QsysResult):
    r"""A result from running an emulator instance.


    Collects data from ``result("tag", val)`` calls in the guppy program. Includes
    results for all shots.

    Includes conversions to traditional distributions over bitstrings if a tagging
    convention is used, including conversion to a pytket BackendResult.

    Under this convention, tags are assumed to be a name of a bit register unless they
    fit the regex pattern ``^([a-z][\w_]*)\[(\d+)\]$`` (like ``my_Reg[12]``) in which
    case they are assumed to refer to the nth element of a bit register.

    For results of the form ``result("<register>", value)``, ``value`` can be ``{0, 1}``
    , wherein the register is assumed to be length 1, or lists over those values,
    wherein the list is taken to be the value of the entire register.

    For results of the form ``result("<register>[n]", value)`` ``value`` can only be
    ``{0,1}``. The register is assumed to be at least `n+1` in size and unset elements
    are assumed to be ``0``.

    Subsequent writes to the same register/element in the same shot will overwrite.

    To convert to a ``BackendResult`` all registers must be present in all shots, and
    register sizes cannot change between shots.

    """

    # cache for extracted partial states, since extraction cleans up the files
    _partial_states: list[list[tuple[str, PartialVector]]] | None = None
    #: List of QsysShot objects, each representing a single shot's results.
    results: list[QsysShot]

    # Re-define parent methods for documentation purposes

    def __init__(self, results: list[QsysShot] | None = None):
        super().__init__(results=results)

    def register_counts(
        self, strict_names: bool = False, strict_lengths: bool = False
    ) -> dict[str, Counter[str]]:
        return super().register_counts(
            strict_names=strict_names, strict_lengths=strict_lengths
        )

    def register_bitstrings(
        self, strict_names: bool = False, strict_lengths: bool = False
    ) -> dict[str, list[str]]:
        return super().register_bitstrings(
            strict_names=strict_names, strict_lengths=strict_lengths
        )

    def to_pytket(self) -> BackendResult:
        return super().to_pytket()

    def collated_shots(self) -> list[dict[str, list[DataValue]]]:
        return super().collated_shots()

    def collated_counts(self) -> Counter[tuple[tuple[str, str], ...]]:
        return super().collated_counts()

    def partial_state_dicts(self) -> list[dict[str, PartialVector]]:
        """Extract state results from shot results in to dictionaries.

        Looks for outputs from `state_result("tag", qs)` calls in the guppy program.

        Returns:
            A list of dictionaries, each dictionary containing the tag as the key and
            the `PartialVector` as the value. Each dictionary corresponds to a shot.
            Repeated tags in a shot will overwrite previous values.
        """
        return [dict(x) for x in self.partial_states()]

    def partial_states(self) -> list[list[tuple[str, PartialVector]]]:
        """Extract state results from shot results.


        Looks for outputs from `state_result("tag", qs)` calls in the guppy program.

        Returns:
            A list (over shots) of lists. The outer list is over shots, and the inner
            is over the state results in that shot.
            Each inner list contains tuples of (string tag, PartialVector).
        """
        if self._partial_states is None:

            def to_partial(
                x: tuple[str, SeleneQuestState],
            ) -> tuple[str, PartialVector]:
                return x[0], PartialVector._from_inner(x[1])

            self._partial_states = [
                list(map(to_partial, Quest.extract_states(shot)))
                for shot in self.results
            ]
        return self._partial_states



from guppylang_internals.experimental import enable_experimental_features

from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std import builtins, debug, quantum
from guppylang.std.builtins import array, comptime, py
from guppylang.std.quantum import qubit

__all__ = (
    "qubit",
    "guppy",
    "enable_experimental_features",
    "GuppyModule",
)

# This is updated by our release-please workflow, triggered by this
# annotation: x-release-please-version
__version__ = "0.21.8"



from typing import Generic, no_type_check

from guppylang_internals.decorator import custom_type, hugr_op
from guppylang_internals.std._internal.compiler.futures import future_op, future_to_hugr
from guppylang_internals.tys.param import TypeParam

from guppylang import guppy
from guppylang.std.lang import owned

T = guppy.type_var("T", copyable=False, droppable=False)

_future_params = [TypeParam(0, "T", must_be_copyable=False, must_be_droppable=False)]


@custom_type(future_to_hugr, copyable=False, droppable=False, params=_future_params)
class Future(Generic[T]):  # type: ignore[misc]
    """A value of type `T` that is computed asynchronously."""

    @hugr_op(future_op("Read"))
    @no_type_check
    def read(self: "Future[T]" @ owned) -> T:
        """Reads a value from a future, consuming it."""

    @hugr_op(future_op("Dup"))
    @no_type_check
    def copy(self: "Future[T]") -> "Future[T]":
        """Duplicate a future."""

    @hugr_op(future_op("Free"))
    @no_type_check
    def discard(self: "Future[T]" @ owned) -> None:
        """Discards a future without reading it."""



"""Guppy standard module for debug functionality."""

# mypy: disable-error-code="empty-body, no-untyped-def"

from guppylang_internals.decorator import custom_function
from guppylang_internals.std._internal.debug import StateResultChecker


@custom_function(checker=StateResultChecker(), higher_order_value=False)
def state_result(tag, *args) -> None:
    """Report the quantum state of the specified qubits.

    This is a debugging function that works only when the program is executed
    on a supported simulator.

    Args:
        tag: A string literal representing the tag of the result.
        args: The qubits whose quantum state is to be reported. The order they are given
        in corresponds to the order in which the state will be reported.
    """



"""
Emulation of Guppy programs powered by the selene-sim package.

Provides a configurable interface for compiling Guppy functions
into an emulator instance, and a configurable builder for setting
instance options and executing.

Emulation returns :py:class:`EmulatorResult` objects, which contain the result output
by the emulation. Results are recorded by calling ``result("tag", value)`` in the
Guppy program, and can include both quantum measurement outcomes and classical
outputs.


Basic emulation
-----------------

Calling ``.emulator`` on a Guppy function compiles it into an
:py:class:`EmulatorInstance`. The method has a required parameter
for the number of qubits to allocate. This cannot be
inferred from the program automatically as it can request an arbitrary number of qubits
at runtime.

Calling ``.run()`` on the instance runs the emulation, returning an
:py:class:`EmulatorResult` object containing the results.

.. code-block:: python

    from guppylang import guppy
    from guppylang.std.builtins import result
    from guppylang.std.quantum import qubit, measure

    @guppy
    def foo() -> None:
        q = qubit()
        result("q", measure(q))

    foo.emulator(n_qubits=1).run()

Change simulation method
--------------------------

The default simulation method is statevector simulation, powered
by the Quest selene plugin.
The simulation method can be changed by calling methods on the
:py:class:`EmulatorBuilder` instance that return updated instances.
For example to use the stabilizer simulator with clifford circuits:

.. code-block:: python

    foo.emulator(1).stabilizer_sim().run()

See also :py:meth:`EmulatorInstance.coinflip_sim` for a no-quantum simulation.

In addition arbitrary selene-sim ``Simulator`` plugins can be used
by calling :py:meth:`EmulatorInstance.with_simulator`.


Configuring emulator instances
-------------------------------

In general the emulation can be configured by chaining ``with_*`` methods on the
:py:class:`EmulatorInstance` object. Each method returns a new instance with the
updated configuration.

For example, the default number of shots run is 1, to change that:

.. code-block:: python

    foo.emulator(n_qubits=1).with_shots(1000).run()

Or update many options at once:

.. code-block:: python

    foo.emulator(n_qubits=1).with_shots(1000).with_seed(42).with_shot_offset(10).run()

See the :py:class:`EmulatorInstance` documentation for a full list of options and their
defaults.


Noisy simulation
-----------------

Selene-sim supports noisy simulation, which can be enabled by setting an error model
on the emulator instance.

.. code-block:: python

    from selene_sim.backends.bundled_error_models import DepolarizingErrorModel

    foo.emulator(1).with_error_model(DepolarizingErrorModel()).run()

State results
-----------------

Calling ``state_result("tag", q1, q2)`` in the Guppy program will record the
state of the qubits `q1` and `q2` in the outputs.
The particular representation of the state depends on the simulation backend,
the default statevector simulator returns a :py:class:`StateVector` object
which is just a numpy array of complex amplitudes.

In general the qubits you request state for may not be all the qubits in the fully
entangled state, in which case the remaining qubits are traced over and a
probabilistic distribution over statws is returned.

The two methods :py:meth:`EmulatorResult.partial_states` and
:meth:`EmulatorResult.partial_state_dicts` extract state results
from the emulator output as :py:class:`PartialVector` objects.

.. code-block:: ipython

    from guppylang import guppy
    from guppylang.std.debug import state_result
    from guppylang.std.quantum import qubit, measure, cx, h

    @guppy
    def foo() -> None:
        # Measure one qubit in a bell state
        q0 = qubit()
        h(q0)
        q1 = qubit()
        cx(q0, q1)
        state_result("q0", q0)
        measure(q0)
        measure(q1)

    res = foo.emulator(2).run()
    # get state outputs at shot 0, tag "q0"
    res.partial_state_dicts()[0]["q0"].state_distribution()

Output is a uniform distribution over the two basis states of the qubit:

.. code-block:: python

    [TracedState(probability=0.5, state=array([1.+0.j, 0.+0.j])),
    TracedState(probability=0.5, state=array([0.+0.j, 1.+0.j]))]
"""

from .builder import EmulatorBuilder
from .exceptions import EmulatorError
from .instance import EmulatorInstance
from .result import EmulatorResult, QsysShot, TaggedResult
from .state import PartialState, PartialVector, StateVector, TracedState

__all__ = [
    "EmulatorBuilder",
    "EmulatorError",
    "EmulatorInstance",
    "EmulatorResult",
    "PartialState",
    "PartialVector",
    "QsysShot",
    "StateVector",
    "TaggedResult",
    "TracedState",
]



"""Operations on booleans."""

# mypy: disable-error-code="empty-body, misc, override, valid-type, no-untyped-def"

from __future__ import annotations

from typing import no_type_check

from guppylang_internals.decorator import custom_function, extend_type, hugr_op
from guppylang_internals.definition.custom import NoopCompiler
from guppylang_internals.std._internal.checker import DunderChecker
from guppylang_internals.std._internal.util import bool_logic_op
from guppylang_internals.tys.builtin import bool_type_def

from guppylang import guppy
from guppylang.std.num import nat


@extend_type(bool_type_def)
class bool:
    """Booleans representing truth values.

    The bool type has exactly two constant instances: ``True`` and ``False``.

    The bool constructor takes a single argument and converts it to ``True`` or
    ``False`` using the standard truth testing procedure.
    """

    @hugr_op(bool_logic_op("and"))
    def __and__(self: bool, other: bool) -> bool: ...

    @custom_function(NoopCompiler())
    def __bool__(self: bool) -> bool: ...

    @hugr_op(bool_logic_op("eq"))
    def __eq__(self: bool, other: bool) -> bool: ...

    @guppy
    @no_type_check
    def __ne__(self: bool, other: bool) -> bool:
        return not self == other

    @guppy
    @no_type_check
    def __int__(self: bool) -> int:
        return 1 if self else 0

    @guppy
    @no_type_check
    def __nat__(self: bool) -> nat:
        # TODO: Type information doesn't flow through the `if` expression, so we
        #  have to insert the `nat` coercions by hand.
        #  See https://github.com/quantinuum/guppylang/issues/707
        return nat(1) if self else nat(0)

    @custom_function(checker=DunderChecker("__bool__"), higher_order_value=False)
    def __new__(x): ...

    @hugr_op(bool_logic_op("or"))
    def __or__(self: bool, other: bool) -> bool: ...

    @hugr_op(bool_logic_op("xor"))
    def __xor__(self: bool, other: bool) -> bool: ...



"""Platform functions."""

# mypy: disable-error-code="empty-body, misc, override, valid-type, no-untyped-def"

from __future__ import annotations

from typing import TYPE_CHECKING, no_type_check

from guppylang_internals.decorator import custom_function
from guppylang_internals.std._internal.checker import (
    BarrierChecker,
    ExitChecker,
    PanicChecker,
)
from guppylang_internals.std._internal.compiler.platform import (
    ArrayResultCompiler,
    ResultCompiler,
)

from guppylang import guppy

if TYPE_CHECKING:
    from guppylang.std.array import array
    from guppylang.std.lang import comptime
    from guppylang.std.num import nat

n = guppy.nat_var("n")


@custom_function(ResultCompiler("result_int", with_int_width=True))
def _result_int(tag: str @ comptime, value: int) -> None: ...


@custom_function(ResultCompiler("result_uint", with_int_width=True))
def _result_nat(tag: str @ comptime, value: nat) -> None: ...


@custom_function(ResultCompiler("result_bool"))
def _result_bool(tag: str @ comptime, value: bool) -> None: ...


@custom_function(ResultCompiler("result_f64"))
def _result_float(tag: str @ comptime, value: float) -> None: ...


@custom_function(ArrayResultCompiler("result_array_int", with_int_width=True))
def _result_int_array(tag: str @ comptime, value: array[int, n]) -> None: ...


@custom_function(ArrayResultCompiler("result_array_uint", with_int_width=True))
def _result_nat_array(tag: str @ comptime, value: array[nat, n]) -> None: ...


@custom_function(ArrayResultCompiler("result_array_bool"))
def _result_bool_array(tag: str @ comptime, value: array[bool, n]) -> None: ...


@custom_function(ArrayResultCompiler("result_array_f64"))
def _result_float_array(tag: str @ comptime, value: array[float, n]) -> None: ...


@guppy.overload(
    _result_int,
    _result_nat,
    _result_bool,
    _result_float,
    _result_int_array,
    _result_nat_array,
    _result_bool_array,
    _result_float_array,
)
def result(tag: str, value):
    """Report a result with the given tag and value.

    This is the primary way to report results from the program back to the user.
    On Quantinuum systems a single shot execution will return a list of pairs of
    (tag, value).

    Args:
        tag: The tag of the result. Must be a string literal
        value: The value of the result. Currently supported value types are `int`,
        `nat`, `float`, and `bool`.
    """


@custom_function(checker=PanicChecker(), higher_order_value=False)
def panic(msg: str, *args):
    """Panic, throwing an error with the given message, and immediately exit the
    program, aborting any subsequent shots.

    Return type is arbitrary, as this function never returns.

    Args:
        msg: The message to display. Must be a string literal.
        args: Arbitrary extra inputs, will not affect the message. Only useful for
        consuming linear values.
    """


@custom_function(checker=ExitChecker(), higher_order_value=False)
def exit(msg: str, signal: int, *args):
    """Exit, reporting the given message and signal, and immediately exit the
    program. Subsequent shots may still run.

    Return type is arbitrary, as this function never returns.

    On Quantinuum systems only signals in the range 1<=signal<=1000 are supported.

    Args:
        msg: The message to display. Must be a string literal.
        args: Arbitrary extra inputs, will not affect the message. Only useful for
        consuming linear values.
    """


@custom_function(checker=BarrierChecker(), higher_order_value=False)
@no_type_check
def barrier(*args) -> None:
    """Barrier to guarantee that all operations before the barrier are completed before
    operations after the barrier are started."""



"""Deprecated: Use `guppylang.std.quantum.functional` instead."""

from guppylang.std.quantum.functional import (
    ch,
    crz,
    cx,
    cy,
    cz,
    h,
    project_z,
    reset,
    rx,
    ry,
    rz,
    s,
    sdg,
    t,
    tdg,
    toffoli,
    v,
    vdg,
    x,
    y,
    z,
)

__all__ = (
    "h",
    "cz",
    "cx",
    "cy",
    "t",
    "s",
    "v",
    "x",
    "y",
    "z",
    "tdg",
    "sdg",
    "vdg",
    "rz",
    "rx",
    "ry",
    "crz",
    "toffoli",
    "reset",
    "project_z",
    "ch",
)



"""Error handling with the `Result` type."""

from typing import Generic, no_type_check

from guppylang_internals.decorator import custom_function, custom_type
from guppylang_internals.definition.custom import NoopCompiler
from guppylang_internals.std._internal.compiler.either import (
    EitherConstructor,
    EitherTestCompiler,
    EitherUnwrapCompiler,
    either_to_hugr,
)
from guppylang_internals.tys.param import TypeParam

from guppylang import guppy
from guppylang.std.either import Either
from guppylang.std.lang import owned

T = guppy.type_var("T", copyable=False, droppable=False)
E = guppy.type_var("E", copyable=False, droppable=False)

_params = [TypeParam(0, "T", False, False), TypeParam(1, "E", False, False)]


@custom_type(either_to_hugr, params=_params)
class Result(Generic[T, E]):  # type: ignore[misc]
    """Represents a union of either an `ok(T)` or an `err(E)` value."""

    @custom_function(EitherTestCompiler(0))
    @no_type_check
    def is_ok(self: "Result[T, E]") -> bool:
        """Returns `True` for an `ok` value."""

    @custom_function(EitherTestCompiler(1))
    @no_type_check
    def is_err(self: "Result[T, E]") -> bool:
        """Returns `True` for an `err` value."""

    @custom_function(EitherUnwrapCompiler(0))
    @no_type_check
    def unwrap(self: "Result[T, E]" @ owned) -> T:
        """Returns the contained `ok` value, consuming `self`.

        Panics if `self` is an `err` value.
        """

    @custom_function(EitherUnwrapCompiler(1))
    @no_type_check
    def unwrap_err(self: "Result[T, E]" @ owned) -> E:
        """Returns the contained `err` value, consuming `self`.

        Panics if `self` is an `ok` value.
        """

    @custom_function(NoopCompiler())
    @no_type_check
    def into_either(self: "Result[T, E]" @ owned) -> Either[T, E]:
        """Casts a `Result` value into an `Either` value."""


@custom_function(EitherConstructor(0))
@no_type_check
def ok(val: T @ owned) -> Result[T, E]:
    """Constructs an `ok` result value."""


@custom_function(EitherConstructor(1))
@no_type_check
def err(err: E @ owned) -> Result[T, E]:
    """Constructs an `err` result value."""



# mypy: disable-error-code="no-any-return"
from typing import Generic, no_type_check

from guppylang_internals.decorator import custom_function, custom_type, hugr_op
from guppylang_internals.std._internal.compiler.qsystem import (
    RandomIntBoundedCompiler,
    RandomIntCompiler,
)
from guppylang_internals.std._internal.compiler.quantum import (
    RNGCONTEXT_T,
)
from guppylang_internals.std._internal.compiler.tket_exts import (
    QSYSTEM_RANDOM_EXTENSION,
)
from guppylang_internals.std._internal.util import external_op

from guppylang import guppy
from guppylang.std.angles import angle, pi
from guppylang.std.builtins import array, owned, panic
from guppylang.std.mem import mem_swap
from guppylang.std.option import Option

SHUFFLE_N = guppy.nat_var("SHUFFLE_N")
SHUFFLE_T = guppy.type_var("SHUFFLE_T", copyable=False, droppable=False)

DISCRETE_N = guppy.nat_var("DISCRETE_N")


@hugr_op(external_op("NewRNGContext", [], ext=QSYSTEM_RANDOM_EXTENSION))
@no_type_check
def _new_rng_context(seed: int) -> Option["RNG"]: ...


@custom_type(RNGCONTEXT_T, copyable=False, droppable=False)
class RNG:
    """Random number generator."""

    @guppy  # type: ignore[misc] # Unsupported decorated constructor type; Self argument missing for a non-static method (or an invalid type for self)
    def __new__(seed: int) -> "RNG":
        """Create a new random number generator using a seed."""
        return _new_rng_context(seed).unwrap()

    @guppy
    def random_angle(self: "RNG") -> angle:
        r"""Generate a random angle in the range :math:`[-\pi, \pi)`."""
        return (2.0 * self.random_float() - 1.0) * pi

    @guppy
    def random_clifford_angle(self: "RNG") -> angle:
        r"""Generate a random Clifford angle (multiple of :math:`\pi/2`)."""
        return self.random_int_bounded(4) * pi / 2

    @hugr_op(external_op("DeleteRNGContext", [], ext=QSYSTEM_RANDOM_EXTENSION))
    @no_type_check
    def discard(self: "RNG" @ owned) -> None:
        """Discard the random number generator."""

    @custom_function(RandomIntCompiler())
    @no_type_check
    def random_int(self: "RNG") -> int:
        """Generate a random 32-bit signed integer."""

    @hugr_op(external_op("RandomFloat", [], ext=QSYSTEM_RANDOM_EXTENSION))
    @no_type_check
    def random_float(self: "RNG") -> float:
        """Generate a random floating point value in the range [0,1)."""

    @custom_function(RandomIntBoundedCompiler())
    @no_type_check
    def random_int_bounded(self: "RNG", bound: int) -> int:
        """Generate a random 32-bit integer in the range [0, bound).

        Args:
            bound: The upper bound of the range, needs to less than 2^31.
        """

    @hugr_op(external_op("RandomAdvance", [], ext=QSYSTEM_RANDOM_EXTENSION))
    @no_type_check
    def random_advance(self: "RNG", delta: int) -> None:
        """Advance or backtrack the RNG state by a given number of steps.

        Args:
            delta: Number of steps to move the RNG state forward (if positive)
            or backward (if negative).
        """

    @guppy
    @no_type_check
    def shuffle(self: "RNG", array: array[SHUFFLE_T, SHUFFLE_N]) -> None:
        """Randomly shuffle the elements of a possibly linear array in place.
        Uses the Fisher-Yates algorithm."""
        for k in range(SHUFFLE_N):
            i = SHUFFLE_N - 1 - k
            j = self.random_int_bounded(i + 1)
            # TODO use array swap once lowering implemented
            # https://github.com/quantinuum/guppylang/issues/924
            if i != j:
                mem_swap(array[i], array[j])


@guppy.struct
class DiscreteDistribution(Generic[DISCRETE_N]):  # type: ignore[misc]
    """A generic probability distribution over a set of the form {0, 1, ..., N-1}.

    Objects of this class should be generated using
    :py:meth:`make_discrete_distribution`.
    """

    # The `sums` array represents the cumulative probability distribution. That is,
    # sums[i] is the probability of drawing a value <= i from the distribution.
    sums: array[float, DISCRETE_N]  # type: ignore[valid-type]

    @guppy
    @no_type_check
    def sample(self: "DiscreteDistribution[DISCRETE_N]", rng: RNG) -> int:
        """Return a sample value from the distribution, using the provided
        :py:class:`RNG`.
        """
        x = rng.random_float()
        # Use binary search to find the least i s.t. sums[i] >= x.
        i_min = 0
        i_max = DISCRETE_N - 1
        while i_min < i_max:
            i = (i_min + i_max) // 2
            if self.sums[i] >= x:
                i_max = i
            else:
                i_min = i + 1
        return i_min


@guppy
@no_type_check
def make_discrete_distribution(
    weights: array[float, DISCRETE_N],
) -> DiscreteDistribution[DISCRETE_N]:
    """Construct a discrete probability distribution over the set
    {0, 1, 2, ... DISCRETE_N - 1}, given as an array of weights which represent
    probabilities. The weights need not be normalized, but must be non-negative and not
    all zero."""
    W = 0.0
    for w in weights.copy():
        if w < 0.0:
            panic("Negative weight included in discrete distribution.")
        W += w
    if W == 0.0:
        panic("No positive weights included in discrete distribution.")
    sums = array(0.0 for _ in range(DISCRETE_N))
    s = 0.0
    for i in range(DISCRETE_N - 1):
        s += weights[i]
        sums[i] = s / W
    sums[DISCRETE_N - 1] = 1.0
    return DiscreteDistribution(sums)



"""Utilities for advanced usage of ownership and borrowing."""

from collections.abc import Callable
from typing import no_type_check

from guppylang_internals.decorator import custom_function
from guppylang_internals.std._internal.compiler.mem import WithOwnedCompiler
from guppylang_internals.std._internal.compiler.prelude import MemSwapCompiler

from guppylang import guppy
from guppylang.std.lang import owned

T = guppy.type_var("T", copyable=False, droppable=False)
Out = guppy.type_var("Out", copyable=False, droppable=False)


@custom_function(MemSwapCompiler())
@no_type_check
def mem_swap(x: T, y: T) -> None:
    """Swaps the values of two variables."""


@custom_function(WithOwnedCompiler())
@no_type_check
def with_owned(val: T, f: Callable[[T @ owned], tuple[Out, T]]) -> Out:
    """Runs a closure where the borrowed argument is promoted to an owned one.

    The closure should return two values:

    * A generic return value that will be passed through.
    * Another value of type `T` that is written back into the borrowed place. This can
      either be the original passed value, or a new value of the same type that was
      created in the closure.

    Pretending that `val` is a pointer, this would be equivalent to the following
    operation in Rust/C: ``(out, *val) = f(*val)``
    """



from collections.abc import Callable
from typing import no_type_check

from guppylang.decorator import guppy
from guppylang.std.builtins import array, comptime, nat

T = guppy.type_var("T", copyable=False, droppable=False)


@guppy
@no_type_check
def spawn_wasm_contexts(n: nat @ comptime, spawn: Callable[[nat], T]) -> "array[T, n]":  # noqa: F821
    return array(spawn(nat(ix)) for ix in range(n))



"""
Compiling and building emulator instances for guppy programs.
"""

from __future__ import annotations

from dataclasses import dataclass, replace
from typing import TYPE_CHECKING, Any

import selene_sim
from typing_extensions import Self

from .instance import EmulatorInstance

if TYPE_CHECKING:
    from collections.abc import Sequence
    from pathlib import Path

    from hugr.package import Package
    from selene_core import BuildPlanner, QuantumInterface, Utility


@dataclass(frozen=True)
class EmulatorBuilder:
    """A builder class for creating EmulatorInstance objects.

    Supports configuration parameters for compilation of emulator instances.
    """

    # interface supported parameters
    _name: str | None = None
    _build_dir: Path | None = None
    _verbose: bool = False

    # selene_sim supported parameters, may be added in the future:
    _planner: BuildPlanner | None = None
    _utilities: Sequence[Utility] | None = None
    _interface: QuantumInterface | None = None
    _progress_bar: bool = False
    _strict: bool = False
    _save_planner: bool = False
    _custom_args: dict[str, Any] | None = None

    @property
    def name(self) -> str | None:
        """User specified name for the emulator instance. Defaults to None."""
        return self._name

    @property
    def build_dir(self) -> Path | None:
        """Directory to store intermediate build files and execution results.
        Defaults to None, in which case a temporary directory is used."""
        return self._build_dir

    @property
    def verbose(self) -> bool:
        """Whether to print verbose output during the build process."""
        return self._verbose

    @property
    def custom_args(self) -> dict[str, Any] | None:
        """Custom build arguments passed to selene_sim.build."""
        return dict(self._custom_args) if self._custom_args is not None else None

    def with_name(self, value: str | None) -> Self:
        """Set the name for the emulator instance."""
        return replace(self, _name=value)

    def with_build_dir(self, value: Path | None) -> Self:
        """Set the build directory for the emulator instance,
        see `EmulatorBuilder.build_dir`."""
        return replace(self, _build_dir=value)

    def with_verbose(self, value: bool) -> Self:
        """Set whether to print verbose output during the build process."""
        return replace(self, _verbose=value)

    def with_build_arg(self, key: str, value: Any) -> Self:
        """Selene builds may support additional customisable arguments,
        e.g. to override the choice of compilation route. This is passed
        to selene_sim.build as kwargs.

        For example:

        .. code-block:: python

            .with_build_arg("build_method", "via-llvm_ir")

        is equivalent to

        .. code-block:: python

            selene_sim.build(..., build_method="via-llvm_ir")

        which saves LLVM IR into the build directory rather than
        saving bitcode.
        """
        if self._custom_args is None:
            return replace(self, _custom_args={key: value})
        else:
            return replace(self, _custom_args=self._custom_args | {key: value})

    def build(self, package: Package, n_qubits: int) -> EmulatorInstance:
        """Build an EmulatorInstance from a compiled package.

        Args:
            package: The compiled HUGR package to build the emulator from.
            n_qubits: The number of qubits to allocate for the emulator instance.

        Returns:
            An EmulatorInstance that can be used to run the compiled program.
        """

        instance = selene_sim.build(  # type: ignore[attr-defined]
            package,
            name=self._name,
            build_dir=self._build_dir,
            interface=self._interface,
            utilities=self._utilities,
            verbose=self._verbose,
            planner=self._planner,
            progress_bar=self._progress_bar,
            strict=self._strict,
            save_planner=self._save_planner,
            **self._custom_args or {},
        )

        return EmulatorInstance(_instance=instance, _n_qubits=n_qubits)



"""Guppy standard module for quantum operations."""

# mypy: disable-error-code="empty-body, misc, valid-type"

from typing import no_type_check

from guppylang_internals.decorator import custom_function, custom_type, hugr_op
from guppylang_internals.std._internal.compiler.quantum import (
    InoutMeasureCompiler,
    RotationCompiler,
)
from guppylang_internals.std._internal.util import quantum_op
from guppylang_internals.tys.ty import UnitaryFlags
from hugr import tys as ht

from guppylang import guppy
from guppylang.std.angles import angle, pi
from guppylang.std.array import array
from guppylang.std.lang import owned
from guppylang.std.option import Option


@custom_type(ht.Qubit, copyable=False, droppable=False)
class qubit:
    @hugr_op(quantum_op("QAlloc"))
    @no_type_check
    def __new__() -> "qubit": ...

    @guppy
    @no_type_check
    def measure(self: "qubit" @ owned) -> bool:
        return measure(self)

    @guppy
    @no_type_check
    def project_z(self: "qubit") -> bool:
        return project_z(self)

    @guppy
    @no_type_check
    def discard(self: "qubit" @ owned) -> None:
        discard(self)


@hugr_op(quantum_op("TryQAlloc"))
@no_type_check
def maybe_qubit() -> Option[qubit]:
    """Try to allocate a qubit, returning `some(qubit)`
    if allocation succeeds or `nothing` if it fails."""


@hugr_op(quantum_op("H"), unitary_flags=UnitaryFlags.Unitary)
@no_type_check
def h(q: qubit) -> None:
    r"""Hadamard gate command

    .. math::
        \mathrm{H}= \frac{1}{\sqrt{2}}
          \begin{pmatrix}
            1 & 1 \\
            1 & -1
          \end{pmatrix}
    """


@hugr_op(quantum_op("CZ"), unitary_flags=UnitaryFlags.Unitary)
@no_type_check
def cz(control: qubit, target: qubit) -> None:
    r"""Controlled-Z gate command.

    cz(control, target)

    Qubit ordering: [control, target]

    .. math::
        \mathrm{CZ}=
          \begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0 \\
            0 & 0 & 1 & 0 \\
            0 & 0 & 0 & -1
          \end{pmatrix}
    """


@hugr_op(quantum_op("CY"), unitary_flags=UnitaryFlags.Unitary)
@no_type_check
def cy(control: qubit, target: qubit) -> None:
    r"""Controlled-Y gate command.

    cy(control, target)

    Qubit ordering: [control, target]

    .. math::
        \mathrm{CY}=
          \begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0 \\
            0 & 0 & 0 & -i \\
            0 & 0 & i & 0
          \end{pmatrix}
    """


@hugr_op(quantum_op("CX"), unitary_flags=UnitaryFlags.Unitary)
@no_type_check
def cx(control: qubit, target: qubit) -> None:
    r"""Controlled-X gate command.

    cx(control, target)

    Qubit ordering: [control, target]

    .. math::
        \mathrm{CX}=
          \begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0 \\
            0 & 0 & 0 & 1 \\
            0 & 0 & 1 & 0
          \end{pmatrix}
    """


@hugr_op(quantum_op("T"), unitary_flags=UnitaryFlags.Unitary)
@no_type_check
def t(q: qubit) -> None:
    r"""T gate.

    .. math::
        \mathrm{T}=
          \begin{pmatrix}
            1 & 0 \\
            0 & e^{i \frac{\pi}{4}}
           \end{pmatrix}

    """


@hugr_op(quantum_op("S"), unitary_flags=UnitaryFlags.Unitary)
@no_type_check
def s(q: qubit) -> None:
    r"""S gate.

    .. math::
        \mathrm{S}=
          \begin{pmatrix}
            1 & 0 \\
            0 & i
           \end{pmatrix}

    """


@hugr_op(quantum_op("V"), unitary_flags=UnitaryFlags.Unitary)
@no_type_check
def v(q: qubit) -> None:
    r"""V gate.

    .. math::
      \mathrm{V}= \frac{1}{\sqrt{2}}
       \begin{pmatrix}
            1 & -i \\
            -i & 1
           \end{pmatrix}

    """


@hugr_op(quantum_op("X"), unitary_flags=UnitaryFlags.Unitary)
@no_type_check
def x(q: qubit) -> None:
    r"""X gate.

    .. math::
        \mathrm{X}=
          \begin{pmatrix}
            0 & 1 \\
            1 & 0
           \end{pmatrix}

    """


@hugr_op(quantum_op("Y"), unitary_flags=UnitaryFlags.Unitary)
@no_type_check
def y(q: qubit) -> None:
    r"""Y gate.

    .. math::
        \mathrm{Y}=
          \begin{pmatrix}
            0 & -i \\
            i & 0
           \end{pmatrix}

    """


@hugr_op(quantum_op("Z"), unitary_flags=UnitaryFlags.Unitary)
@no_type_check
def z(q: qubit) -> None:
    r"""Z gate.

    .. math::
        \mathrm{Z}=
          \begin{pmatrix}
            1 & 0 \\
            0 & -1
           \end{pmatrix}

    """


@hugr_op(quantum_op("Tdg"), unitary_flags=UnitaryFlags.Unitary)
@no_type_check
def tdg(q: qubit) -> None:
    r"""Tdg gate.

    .. math::
        \mathrm{T}^\dagger=
          \begin{pmatrix}
            1 & 0 \\
            0 & e^{-i \frac{\pi}{4}}
           \end{pmatrix}

    """


@hugr_op(quantum_op("Sdg"), unitary_flags=UnitaryFlags.Unitary)
@no_type_check
def sdg(q: qubit) -> None:
    r"""Sdg gate.

    .. math::
        \mathrm{S}^\dagger=
          \begin{pmatrix}
            1 & 0 \\
            0 & -i
           \end{pmatrix}

    """


@hugr_op(quantum_op("Vdg"), unitary_flags=UnitaryFlags.Unitary)
@no_type_check
def vdg(q: qubit) -> None:
    r"""Vdg gate.

    .. math::
      \mathrm{V}^\dagger= \frac{1}{\sqrt{2}}
       \begin{pmatrix}
            1 & i \\
            i & 1
           \end{pmatrix}

    """


@custom_function(RotationCompiler("Rz"), unitary_flags=UnitaryFlags.Unitary)
@no_type_check
def rz(q: qubit, angle: angle) -> None:
    r"""Rz gate.

    .. math::
        \mathrm{Rz}(\theta)=
        \exp(\frac{- i  \theta}{2} Z)=
          \begin{pmatrix}
            e^{-\frac{1}{2}i  \theta} & 0 \\
            0 & e^{\frac{1}{2}i  \theta}
           \end{pmatrix}

    """


@custom_function(RotationCompiler("Rx"), unitary_flags=UnitaryFlags.Unitary)
@no_type_check
def rx(q: qubit, angle: angle) -> None:
    r"""Rx gate.

    .. math::
        \mathrm{Rx}(\theta)=
          \begin{pmatrix}
            \cos(\frac{ \theta}{2}) & -i\sin(\frac{ \theta}{2}) \\
            -i\sin(\frac{ \theta}{2}) & \cos(\frac{ \theta}{2})
           \end{pmatrix}

    """


@custom_function(RotationCompiler("Ry"), unitary_flags=UnitaryFlags.Unitary)
@no_type_check
def ry(q: qubit, angle: angle) -> None:
    r"""Ry gate.

    .. math::
        \mathrm{Ry}(\theta)=
          \begin{pmatrix}
            \cos(\frac{\theta}{2}) & -\sin(\frac{ \theta}{2}) \\
            \sin(\frac{ \theta}{2}) & \cos(\frac{ \theta}{2})
           \end{pmatrix}

    """


@custom_function(RotationCompiler("CRz"), unitary_flags=UnitaryFlags.Unitary)
@no_type_check
def crz(control: qubit, target: qubit, angle: angle) -> None:
    r"""Controlled-Rz gate command.

    crz(control, target, theta)

    Qubit ordering: [control, target]

    .. math::
        \mathrm{CRz}(\theta)=
          \begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0 \\
            0 & 0 & e^{-\frac{1}{2}i  \theta} & 0 \\
            0 & 0 & 0 & e^{\frac{1}{2}i  \theta}
        \end{pmatrix}
    """


@hugr_op(quantum_op("Toffoli"), unitary_flags=UnitaryFlags.Unitary)
@no_type_check
def toffoli(control1: qubit, control2: qubit, target: qubit) -> None:
    r"""A Toffoli gate command. Also sometimes known as a CCX gate.

    toffoli(control1, control2, target)

    Qubit ordering: [control1, control2 target]

    .. math::
        \mathrm{Toffoli}=
          \begin{pmatrix}
            1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 1 \\
            0 & 0 & 0 & 0 & 0 & 0 & 1 & 0
          \end{pmatrix}
    """


@custom_function(InoutMeasureCompiler())
@no_type_check
def project_z(q: qubit) -> bool:
    """Project a single qubit into the Z-basis (a non-destructive measurement)."""


@hugr_op(quantum_op("QFree"))
@no_type_check
def discard(q: qubit @ owned) -> None:
    """Discard a single qubit."""


@hugr_op(quantum_op("MeasureFree"))
@no_type_check
def measure(q: qubit @ owned) -> bool:
    """Measure a single qubit destructively."""


@hugr_op(quantum_op("Reset"))
@no_type_check
def reset(q: qubit) -> None:
    """Reset a single qubit to the :math:`|0\rangle` state."""


N = guppy.nat_var("N")


@guppy
@no_type_check
def measure_array(qubits: array[qubit, N] @ owned) -> array[bool, N]:
    """Measure an array of qubits, returning an array of bools."""
    return array(measure(q) for q in qubits)


@guppy
@no_type_check
def discard_array(qubits: array[qubit, N] @ owned) -> None:
    """Discard an array of qubits."""
    for q in qubits:
        discard(q)


# -------NON-PRIMITIVE-------


@guppy
@no_type_check
def ch(control: qubit, target: qubit) -> None:
    r"""Controlled-H gate command.

    ch(control, target)

    Qubit ordering: [control, target]

    .. math::
        \mathrm{CH} = \frac{1}{\sqrt{2}}
          \begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0 \\
            0 & 0 & 1 & 1 \\
            0 & 0 & 1 & -1
        \end{pmatrix}
    """
    # based on https://quantumcomputing.stackexchange.com/a/15737
    ry(target, -pi / 4)
    cz(control, target)
    ry(target, pi / 4)



"""Reexports core types and functions that are available without an explicit import."""

from guppylang.std.array import ArrayIter, FrozenarrayIter, array, frozenarray
from guppylang.std.bool import bool
from guppylang.std.iter import Range, SizedIter, range
from guppylang.std.lang import comptime, owned, py
from guppylang.std.list import list
from guppylang.std.mem import mem_swap
from guppylang.std.num import (
    abs,
    bytecast_float_to_nat,
    bytecast_nat_to_float,
    divmod,
    float,
    int,
    len,
    nat,
    pow,
    round,
)
from guppylang.std.option import Option, nothing, some
from guppylang.std.platform import barrier, exit, panic, result
from guppylang.std.quantum import qubit
from guppylang.std.reflection import callable
from guppylang.std.string import str
from guppylang.std.unsupported import (
    __import__,
    aiter,
    all,
    anext,
    any,
    bin,
    breakpoint,
    bytearray,
    bytes,
    chr,
    classmethod,
    compile,
    complex,
    delattr,
    dict,
    dir,
    enumerate,
    eval,
    exec,
    filter,
    format,
    frozenset,
    getattr,
    globals,
    hasattr,
    hash,
    help,
    hex,
    id,
    input,
    isinstance,
    issubclass,
    iter,
    locals,
    map,
    max,
    memoryview,
    min,
    next,
    object,
    oct,
    open,
    ord,
    print,
    property,
    repr,
    reversed,
    set,
    setattr,
    slice,
    sorted,
    staticmethod,
    sum,
    super,
    type,
    vars,
    zip,
)

__all__ = (
    "__import__",
    "abs",
    "aiter",
    "all",
    "anext",
    "any",
    "array",
    "bin",
    "bool",
    "breakpoint",
    "bytearray",
    "bytes",
    "callable",
    "chr",
    "classmethod",
    "compile",
    "complex",
    "comptime",
    "delattr",
    "dict",
    "dir",
    "divmod",
    "enumerate",
    "eval",
    "exec",
    "exit",
    "filter",
    "float",
    "format",
    "frozenarray",
    "frozenset",
    "getattr",
    "globals",
    "hasattr",
    "hash",
    "help",
    "hex",
    "id",
    "input",
    "int",
    "isinstance",
    "issubclass",
    "iter",
    "len",
    "list",
    "locals",
    "map",
    "max",
    "memoryview",
    "min",
    "nat",
    "next",
    "nothing",
    "object",
    "oct",
    "open",
    "Option",
    "ord",
    "owned",
    "panic",
    "pow",
    "print",
    "property",
    "py",
    "qubit",
    "range",
    "repr",
    "result",
    "reversed",
    "round",
    "set",
    "setattr",
    "SizedIter",
    "slice",
    "some",
    "sorted",
    "staticmethod",
    "str",
    "sum",
    "super",
    "type",
    "vars",
    "zip",
    # TODO: Remove the following from prelude
    "ArrayIter",  # Deprecated reexport
    "barrier",  # Deprecated reexport
    "bytecast_float_to_nat",  # Deprecated reexport
    "bytecast_nat_to_float",  # Deprecated reexport
    "FrozenarrayIter",  # Deprecated reexport
    "mem_swap",  # Deprecated reexport
    "Range",  # Deprecated reexport
)



from typing import no_type_check

from guppylang_internals.decorator import custom_function, hugr_op
from guppylang_internals.definition.custom import BoolOpCompiler
from guppylang_internals.std._internal.compiler.quantum import (
    InoutMeasureCompiler,
    InoutMeasureResetCompiler,
)
from guppylang_internals.std._internal.compiler.tket_exts import QSYSTEM_EXTENSION
from guppylang_internals.std._internal.util import quantum_op

from guppylang import guppy
from guppylang.std import angles
from guppylang.std.angles import angle, pi
from guppylang.std.builtins import owned
from guppylang.std.futures import Future
from guppylang.std.option import Option, nothing, some
from guppylang.std.quantum import qubit


@guppy
@no_type_check
def phased_x(q: qubit, angle1: angle, angle2: angle) -> None:
    r"""phased_x gate command.

    .. math::

        \mathrm{PhasedX}(\theta_1, \theta_2)=
          \mathrm{Rz(\theta_2)Rx(\theta_1)Rz(-\theta_2)}&=
          \begin{pmatrix}
          \cos(\frac{ \theta_1}{2}) &
            -i e^{-i \theta_2}\sin(\frac{\theta_1}{2})\\
          -i e^{i \theta_2}\sin(\frac{\theta_1}{2}) &
            \cos(\frac{\theta_1}{2})
           \end{pmatrix}
    """
    f1 = float(angle1)
    f2 = float(angle2)
    _phased_x(q, f1, f2)


@guppy
@no_type_check
def zz_max(q1: qubit, q2: qubit) -> None:
    r"""zz_max gate command. A maximally entangling zz_phase gate.

    This is a special case of the zz_phase command with :math:`\theta = \frac{\pi}{2}`.

    zz_max(q1, q2)

    Qubit ordering: [q1, q2]

    .. math::
        \mathrm{ZZMax}=
        \exp(\frac{- i\pi}{4}\big(Z \otimes Z \big))=
          \begin{pmatrix}
            e^{\frac{-i\pi}{4}} & 0 & 0 & 0 \\
            0 & e^{\frac{i\pi}{4}} & 0 & 0 \\
            0 & 0 & e^{\frac{i\pi}{4}} & 0 \\
            0 & 0 & 0 & e^{\frac{-i\pi}{4}}
        \end{pmatrix}
    """
    zz_phase(q1, q2, pi / 2)


@guppy
@no_type_check
def zz_phase(q1: qubit, q2: qubit, angle: angle) -> None:
    r"""zz_phase gate command.

    zz_phase(q1, q2, theta)

    Qubit ordering: [q1, q2]

    .. math::
        \mathrm{ZZPhase}(\theta)=
        \exp(\frac{- i \theta}{2}\big(Z \otimes Z \big))=
          \begin{pmatrix}
            e^{\frac{-i \theta}{2}} & 0 & 0 & 0 \\
            0 & e^{\frac{i \theta}{2}} & 0 & 0 \\
            0 & 0 & e^{\frac{i \theta}{2}} & 0 \\
            0 & 0 & 0 & e^{\frac{-i \theta}{2}}
        \end{pmatrix}

    >>> @guppy
    ... def qsystem_cx(q1: qubit, q2: qubit) -> None:
    ...     phased_x(angle(3/2), angle(-1/2), q2)
    ...     zz_phase(q1, q2, angle(1/2))
    ...     rz(angle(5/2), q1)
    ...     phasedx(angle(-3/2), q1)
    ...     rz(angle(3/2), q2)
    >>> qsystem_cx.compile()
    """
    f = float(angle)
    _zz_phase(q1, q2, f)


@guppy
@no_type_check
def rz(q: qubit, angle: angle) -> None:
    r"""rz gate command.

    .. math::
        \mathrm{Rz}(\theta)=
        \exp(\frac{- i \theta}{2} Z)=
          \begin{pmatrix}
            e^{\frac{-i \theta}{2}} & 0  \\
            0 & e^{\frac{i \theta}{2}}
        \end{pmatrix}
    """

    f1 = float(angle)
    _rz(q, f1)


@hugr_op(quantum_op("Measure", ext=QSYSTEM_EXTENSION))
@no_type_check
def measure(q: qubit @ owned) -> bool:
    """Measure a qubit destructively."""


@custom_function(InoutMeasureResetCompiler("MeasureReset", QSYSTEM_EXTENSION))
@no_type_check
def measure_and_reset(q: qubit) -> bool:
    """MeasureReset operation from the qsystem extension."""


@hugr_op(quantum_op("Reset", ext=QSYSTEM_EXTENSION))
@no_type_check
def reset(q: qubit) -> None:
    """Reset a qubit to the :math:`|0\\rangle` state."""


# TODO
# @hugr_op(quantum_op("TryQAlloc", ext=QSYSTEM_EXTENSION))
# @no_type_check
# def _try_qalloc() -> Option[qubit]:
#     ..


@hugr_op(quantum_op("QFree", ext=QSYSTEM_EXTENSION))
@no_type_check
def qfree(q: qubit @ owned) -> None: ...


@hugr_op(quantum_op("LazyMeasureLeaked", ext=QSYSTEM_EXTENSION))
@no_type_check
def _measure_leaked(q: qubit @ owned) -> Future[int]:
    """Measure the quibit or return 2 if it is leaked."""


@guppy
@no_type_check
def measure_leaked(q: qubit @ owned) -> "MaybeLeaked":
    """Measure the qubit and return a MaybeLeaked result."""
    fm = _measure_leaked(q)
    return MaybeLeaked(fm)


@guppy.struct
@no_type_check
class MaybeLeaked:
    """A class representing a measurement that may have leaked.

    This is used to represent the result of `measure_leaked`, which can either
    return a boolean measurement result or indicate that the qubit has leaked.
    """

    _measurement: Future[int]  # type: ignore[type-arg]

    @guppy
    @no_type_check
    def is_leaked(self: "MaybeLeaked") -> bool:
        """Check if the measurement indicates a leak."""
        return self._measurement.copy().read() == 2

    @guppy
    @no_type_check
    def to_result(self: "MaybeLeaked @ owned") -> Option[bool]:
        """Returns the measurement result or `nothing` if leaked."""
        int_value: int = self._measurement.read()
        if int_value == 2:
            return nothing()
        measurement = int_value == 1
        return some(measurement)

    @guppy
    @no_type_check
    def discard(self: "MaybeLeaked @ owned") -> None:
        self._measurement.discard()


# ------------------------------------------------------
# --------- Internal definitions -----------------------
# ------------------------------------------------------


@hugr_op(quantum_op("PhasedX", ext=QSYSTEM_EXTENSION))
@no_type_check
def _phased_x(q: qubit, angle1: float, angle2: float) -> None:
    """PhasedX operation from the qsystem extension.

    See ``guppylang.std.qsystem.phased_x`` for a public definition that
    accepts angle parameters.
    """


@hugr_op(quantum_op("ZZPhase", ext=QSYSTEM_EXTENSION))
@no_type_check
def _zz_phase(q1: qubit, q2: qubit, angle: float) -> None:
    """ZZPhase operation from the qsystem extension.

    See ``guppylang.std.qsystem.zz_phase`` for a public definition that
    accepts angle parameters.
    """


@hugr_op(quantum_op("Rz", ext=QSYSTEM_EXTENSION))
@no_type_check
def _rz(q: qubit, angle: float) -> None:
    """Rz operation from the qsystem extension.

    See ``guppylang.std.qsystem.rz`` for a public definition that
    accepts angle parameters.
    """



"""Definition objects that are being exposed to users.

These are the objects returned by the `@guppy` decorator. They should not be confused
with the compiler-internal definition objects in the `definitions` module.
"""

import ast
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any, ClassVar, Generic, ParamSpec, TypeVar, cast

import guppylang_internals
from guppylang_internals.definition.function import RawFunctionDef
from guppylang_internals.definition.value import CompiledCallableDef
from guppylang_internals.diagnostic import Error, Note
from guppylang_internals.engine import ENGINE, CoreMetadataKeys
from guppylang_internals.error import GuppyError, pretty_errors
from guppylang_internals.span import Span, to_span
from guppylang_internals.tracing.object import TracingDefMixin
from guppylang_internals.tracing.util import hide_trace
from hugr.hugr import Hugr
from hugr.package import Package

import guppylang
from guppylang.emulator import EmulatorBuilder, EmulatorInstance
from guppylang.emulator.exceptions import EmulatorBuildError

__all__ = ("GuppyDefinition", "GuppyFunctionDefinition", "GuppyTypeVarDefinition")


P = ParamSpec("P")
Out = TypeVar("Out")


def _update_generator_metadata(hugr: Hugr[Any]) -> None:
    """Update the generator metadata of a Hugr to be
    guppylang rather than just internals."""
    key = CoreMetadataKeys.GENERATOR.value

    hugr.module_root.metadata[key] = {
        "name": f"guppylang (guppylang-internals-v{guppylang_internals.__version__})",
        "version": guppylang.__version__,
    }


@dataclass(frozen=True)
class EntrypointArgsError(Error):
    title: ClassVar[str] = "Entrypoint function has arguments"
    span_label: ClassVar[str] = (
        "Entrypoint function must have no input parameters, found ({input_names})."
    )
    args: Sequence[str]

    @dataclass(frozen=True)
    class AlternateHint(Note):
        message: ClassVar[str] = (
            "If the function is not an execution entrypoint,"
            " consider using `{function_name}.compile_function()`"
        )
        function_name: str

    @property
    def input_names(self) -> str:
        """Returns a comma-separated list of input names."""
        return ", ".join(f"`{x}`" for x in self.args)


@dataclass(frozen=True)
class GuppyDefinition(TracingDefMixin):
    """A general Guppy definition."""

    def compile(self) -> Package:
        """Compile a Guppy definition to HUGR."""
        package: Package = ENGINE.compile(self.id).package
        for mod in package.modules:
            _update_generator_metadata(mod)
        return package

    def check(self) -> None:
        """Type-check a Guppy definition."""
        return ENGINE.check(self.id)


@dataclass(frozen=True)
class GuppyFunctionDefinition(GuppyDefinition, Generic[P, Out]):
    """A Guppy function definition."""

    @hide_trace
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Out:
        return cast(Out, super().__call__(*args, **kwargs))

    def emulator(
        self, n_qubits: int | None = None, builder: EmulatorBuilder | None = None
    ) -> EmulatorInstance:
        """Compile this function for emulation with the selene-sim emulator.

        Calls `compile()` to get the HUGR package and then builds it using the
        provided `EmulatorBuilder` configuration or a default one.

        See :py:mod:`guppylang.emulator` for more details on the emulator.


        Args:
            n_qubits: The number of qubits to allocate for the function. If it is not
            provided, the function has to declare the maximum number of qubits it needs
            in the decorator, e.g. `@guppy(max_qubits=5)`.
            builder: An optional `EmulatorBuilder` to use for building the emulator
            instance. If not provided, the default `EmulatorBuilder` will be used.

        Returns:
            An `EmulatorInstance` that can be used to run the function in an emulator.
        """
        mod = self.compile()

        builder = builder or EmulatorBuilder()
        qubits = n_qubits
        if (
            isinstance(self.wrapped, RawFunctionDef)
            and self.wrapped.metadata is not None
        ):
            hinted_qubits = self.wrapped.metadata.max_qubits.value
            if qubits is None:
                qubits = hinted_qubits
            elif hinted_qubits is not None and qubits < hinted_qubits:
                raise EmulatorBuildError(
                    ValueError(
                        f"Number of qubits requested ({qubits}) is insufficient to "
                        "cover the maximum number of qubits hinted on the "
                        f"entrypoint ({hinted_qubits})."
                    )
                )

        if qubits is None:
            raise EmulatorBuildError(
                ValueError(
                    "Number of qubits to be used must be specified, either as an "
                    f"argument to `{self.emulator.__name__}` or hinted on the "
                    "entrypoint function using `@guppy(max_qubits=...)`."
                )
            )

        return builder.build(mod, n_qubits=qubits)

    def compile(self) -> Package:
        """
        Compiles an execution entrypoint function definition to a HUGR package

        Equivalent to :py:meth:`GuppyDefinition.compile_entrypoint`.


        Returns:
            Package: The compiled package object.
        Raises:
            GuppyError: If the entrypoint has arguments.
        """

        return self.compile_entrypoint()

    @pretty_errors
    def compile_entrypoint(self) -> Package:
        """
        Compiles an execution entrypoint function definition to a HUGR package

        Returns:
            Package: The compiled package object.
        Raises:
            GuppyError: If the entrypoint has arguments.
        """

        pack = self.compile_function()
        # entrypoint cannot be polymorphic
        monomorphized_id = (self.id, ())
        compiled_def = ENGINE.compiled.get(monomorphized_id)
        if (
            isinstance(compiled_def, CompiledCallableDef)
            and len(compiled_def.ty.inputs) > 0
        ):
            # Check if the entrypoint has arguments
            defined_at = cast(ast.FunctionDef, compiled_def.defined_at)
            start = to_span(defined_at.args.args[0])
            end = to_span(defined_at.args.args[-1])
            span = Span(start=start.start, end=end.end)
            raise GuppyError(
                EntrypointArgsError(
                    span=span,
                    args=compiled_def.ty.input_names or "",
                ).add_sub_diagnostic(
                    EntrypointArgsError.AlternateHint(
                        None, function_name=defined_at.name
                    )
                )
            )
        return pack

    def compile_function(self) -> Package:
        """Compile a Guppy function definition to HUGR.


        Returns:
            Package: The compiled package object.
        """
        return super().compile()


@dataclass(frozen=True)
class GuppyTypeVarDefinition(GuppyDefinition):
    """Definition of a Guppy type variable."""

    # For type variables, we need a `GuppyDefinition` subclass that answers 'yes' to an
    # instance check on `typing.TypeVar`. This hack is needed since `typing.Generic[T]`
    # has a runtime check that enforces that the passed `T` is actually a `TypeVar`.

    __class__: ClassVar[type] = TypeVar

    _ty_var: TypeVar

    def __eq__(self, other: object) -> bool:
        # We need to compare as equal to an equivalent regular type var
        if isinstance(other, TypeVar):
            return self._ty_var == other
        return object.__eq__(self, other)

    def __getattr__(self, name: str) -> Any:
        # Pretend to be a `TypeVar` by providing all of its attributes
        if hasattr(self._ty_var, name):
            return getattr(self._ty_var, name)
        return object.__getattribute__(self, name)



from __future__ import annotations

from typing import Generic, no_type_check

from guppylang.decorator import guppy
from guppylang.std.builtins import array, owned, panic
from guppylang.std.option import Option, nothing, some

T = guppy.type_var("T", copyable=False, droppable=False)
TCopyable = guppy.type_var("TCopyable", copyable=True, droppable=False)
MAX_SIZE = guppy.nat_var("MAX_SIZE")


@guppy.struct
class PriorityQueue(Generic[T, MAX_SIZE]):  # type: ignore[misc]
    """A queue of values ordered by priority.

    Values with the lowest priority value will be popped from the queue first.
    (e.g. Priority 1 will be returned before priority 2.) To ensure static
    allocation, the maximum queue size must be specified in advance and
    is tracked in the type. For example, the `PriorityQueue[int, 10]`
    is a queue that can hold at most 10 prioritized integers.

    Use `empty_priority_queue` to construct a new priority_queue.
    """

    #: Underlying buffer holding the priority queue elements.
    #:
    #: INVARIANT: All array elements up to and including index `self.size - 1` are
    #: `option.some` variants and all further ones are `option.nothing`.
    buf: array[Option[tuple[int, T]], MAX_SIZE]  # type: ignore[valid-type, type-arg]

    #: Index of the next free index in `self.buf`.
    size: int

    @guppy
    @no_type_check
    def __len__(self: PriorityQueue[T, MAX_SIZE]) -> int:
        """Returns the number of elements currently stored in the priority queue."""
        return self.size

    @guppy
    @no_type_check
    def __iter__(
        self: PriorityQueue[T, MAX_SIZE] @ owned,
    ) -> PriorityQueue[T, MAX_SIZE]:
        """Returns an iterator over the queued elements paired with their priority.

        Elements are yielded in priority order.
        """
        return self

    @guppy
    @no_type_check
    def __next__(
        self: PriorityQueue[T, MAX_SIZE] @ owned,
    ) -> Option[tuple[tuple[int, T], PriorityQueue[T, MAX_SIZE]]]:
        if len(self) == 0:
            self.discard_empty()
            return nothing()
        prio, val, new_queue = self.pop()
        return some(((prio, val), new_queue))

    @guppy
    @no_type_check
    def push(
        self: PriorityQueue[T, MAX_SIZE] @ owned,
        value: T @ owned,
        priority: int,
    ) -> PriorityQueue[T, MAX_SIZE]:
        """Adds an element in the correct order to the priority queue.

        Panics if the priority queue has already reached its maximum size.
        """
        if self.size >= MAX_SIZE:
            panic("PriorityQueue.push: max size reached")
        self.buf[self.size].swap(some((priority, value))).unwrap_nothing()
        i = self.size
        while i > 0:
            parent_i = (i - 1) // 2
            prio, val = self.buf[i].take().unwrap()
            parent_prio, parent_val = self.buf[parent_i].take().unwrap()
            if prio >= parent_prio:
                self.buf[i].swap(some((prio, val))).unwrap_nothing()
                self.buf[parent_i].swap(
                    some((parent_prio, parent_val))
                ).unwrap_nothing()
                break
            self.buf[i].swap(some((parent_prio, parent_val))).unwrap_nothing()
            self.buf[parent_i].swap(some((prio, val))).unwrap_nothing()
            i = parent_i
        return PriorityQueue(self.buf, self.size + 1)

    @guppy
    @no_type_check
    def pop(
        self: PriorityQueue[T, MAX_SIZE] @ owned,
    ) -> tuple[int, T, PriorityQueue[T, MAX_SIZE]]:
        """Removes the next element from the priority queue.

        Panics if the priority queue is empty.
        """
        if self.size <= 0:
            panic("PriorityQueue.pop: priority queue is empty")
        return_prio, return_val = self.buf[0].take().unwrap()
        new_size = self.size - 1
        if new_size == 0:
            return return_prio, return_val, PriorityQueue(self.buf, new_size)
        displaced_prio, displaced_val = self.buf[new_size].take().unwrap()
        i = 0
        while True:
            left_i = 2 * i + 1
            if left_i >= new_size:
                break
            right_i = left_i + 1
            if right_i < new_size:
                left_elem = self.buf[left_i].take().unwrap()
                right_elem = self.buf[right_i].take().unwrap()
                left_prio, left_val = left_elem
                right_prio, right_val = right_elem
                if right_prio < left_prio:
                    child_i, child_prio, child_val = right_i, right_prio, right_val
                    self.buf[left_i].swap(some((left_prio, left_val))).unwrap_nothing()
                else:
                    child_i, child_prio, child_val = left_i, left_prio, left_val
                    self.buf[right_i].swap(
                        some((right_prio, right_val))
                    ).unwrap_nothing()
            else:
                left_elem = self.buf[left_i].take().unwrap()
                child_prio, child_val = left_elem
                child_i = left_i
            if displaced_prio <= child_prio:
                self.buf[child_i].swap(some((child_prio, child_val))).unwrap_nothing()
                break
            self.buf[i].swap(some((child_prio, child_val))).unwrap_nothing()
            i = child_i
        self.buf[i].swap(some((displaced_prio, displaced_val))).unwrap_nothing()
        return return_prio, return_val, PriorityQueue(self.buf, new_size)

    @guppy
    @no_type_check
    def peek(
        self: PriorityQueue[TCopyable, MAX_SIZE] @ owned,
    ) -> tuple[int, TCopyable, PriorityQueue[TCopyable, MAX_SIZE]]:
        """Returns a copy of the next element in the priority queue without removing it.

        Panics if the priority queue is empty.

        Note that this operation is only allowed if the priority queue elements are
        copyable.
        """
        if self.size <= 0:
            panic("PriorityQueue.peek: priority queue is empty")
        prio, val = self.buf[0].unwrap()
        return prio, val, PriorityQueue(self.buf, self.size)

    @guppy
    @no_type_check
    def discard_empty(self: PriorityQueue[T, MAX_SIZE] @ owned) -> None:
        """Discards a priority queue of potentially non-droppable elements assuming that
        the queue is empty.

        Panics if the queue is not empty.
        """
        if self.size > 0:
            panic("PriorityQueue.discard_empty: priority queue is not empty")
        for elem in self.buf:
            elem.unwrap_nothing()


@guppy
@no_type_check
def empty_priority_queue() -> PriorityQueue[T, MAX_SIZE]:
    """Constructs a new empty priority queue."""
    buf = array(nothing[tuple[int, T]]() for _ in range(MAX_SIZE))  # type: ignore[valid-type]
    return PriorityQueue(buf, 0)



"""Guppy standard module for functional quantum operations. For the mathematical
definitions of these gates, see the guppylang.std.quantum documentation.

These gates are the same as those in std.quantum but use functional syntax.
"""

from typing import no_type_check

import guppylang.std.quantum as quantum
from guppylang.decorator import guppy

# mypy: disable-error-code="empty-body, misc, valid-type"
from guppylang.std.angles import angle
from guppylang.std.lang import owned
from guppylang.std.quantum import qubit


@guppy
@no_type_check
def h(q: qubit @ owned) -> qubit:
    """Functional Hadamard gate command."""
    quantum.h(q)
    return q


@guppy
@no_type_check
def cz(control: qubit @ owned, target: qubit @ owned) -> tuple[qubit, qubit]:
    """Functional CZ gate command."""
    quantum.cz(control, target)
    return control, target


@guppy
@no_type_check
def cx(control: qubit @ owned, target: qubit @ owned) -> tuple[qubit, qubit]:
    """Functional CX gate command."""
    quantum.cx(control, target)
    return control, target


@guppy
@no_type_check
def cy(control: qubit @ owned, target: qubit @ owned) -> tuple[qubit, qubit]:
    """Functional CY gate command."""
    quantum.cy(control, target)
    return control, target


@guppy
@no_type_check
def t(q: qubit @ owned) -> qubit:
    """Functional T gate command."""
    quantum.t(q)
    return q


@guppy
@no_type_check
def s(q: qubit @ owned) -> qubit:
    """Functional S gate command."""
    quantum.s(q)
    return q


@guppy
@no_type_check
def v(q: qubit @ owned) -> qubit:
    """Functional V gate command."""
    quantum.v(q)
    return q


@guppy
@no_type_check
def x(q: qubit @ owned) -> qubit:
    """Functional X gate command."""
    quantum.x(q)
    return q


@guppy
@no_type_check
def y(q: qubit @ owned) -> qubit:
    """Functional Y gate command."""
    quantum.y(q)
    return q


@guppy
@no_type_check
def z(q: qubit @ owned) -> qubit:
    """Functional Z gate command."""
    quantum.z(q)
    return q


@guppy
@no_type_check
def tdg(q: qubit @ owned) -> qubit:
    """Functional Tdg gate command."""
    quantum.tdg(q)
    return q


@guppy
@no_type_check
def sdg(q: qubit @ owned) -> qubit:
    """Functional Sdg gate command."""
    quantum.sdg(q)
    return q


@guppy
@no_type_check
def vdg(q: qubit @ owned) -> qubit:
    """Functional Vdg gate command."""
    quantum.vdg(q)
    return q


@guppy
@no_type_check
def rz(q: qubit @ owned, angle: angle) -> qubit:
    """Functional Rz gate command."""
    quantum.rz(q, angle)
    return q


@guppy
@no_type_check
def rx(q: qubit @ owned, angle: angle) -> qubit:
    """Functional Rx gate command."""
    quantum.rx(q, angle)
    return q


@guppy
@no_type_check
def ry(q: qubit @ owned, angle: angle) -> qubit:
    """Functional Ry gate command."""
    quantum.ry(q, angle)
    return q


@guppy
@no_type_check
def crz(
    control: qubit @ owned, target: qubit @ owned, angle: angle
) -> tuple[qubit, qubit]:
    """Functional CRz gate command."""
    quantum.crz(control, target, angle)
    return control, target


@guppy
@no_type_check
def toffoli(
    control1: qubit @ owned, control2: qubit @ owned, target: qubit @ owned
) -> tuple[qubit, qubit, qubit]:
    """Functional Toffoli gate command."""
    quantum.toffoli(control1, control2, target)
    return control1, control2, target


@guppy
@no_type_check
def reset(q: qubit @ owned) -> qubit:
    """Functional Reset command."""
    quantum.reset(q)
    return q


@guppy
@no_type_check
def project_z(q: qubit @ owned) -> tuple[qubit, bool]:
    """Functional project_z command."""
    b = quantum.project_z(q)
    return q, b


# -------NON-PRIMITIVE-------


@guppy
@no_type_check
def ch(control: qubit @ owned, target: qubit @ owned) -> tuple[qubit, qubit]:
    """Functional Controlled-H gate command."""
    quantum.ch(control, target)
    return control, target



import ast
import builtins
import inspect
from collections.abc import Callable, Sequence
from types import FrameType
from typing import Any, ParamSpec, TypedDict, TypeVar, cast, overload

from guppylang_internals.ast_util import annotate_location
from guppylang_internals.compiler.core import (
    CompilerContext,
)
from guppylang_internals.decorator import (
    custom_function,
    custom_type,
    hugr_op,
)
from guppylang_internals.definition.common import DefId
from guppylang_internals.definition.const import RawConstDef
from guppylang_internals.definition.custom import (
    CustomCallChecker,
    CustomInoutCallCompiler,
    RawCustomFunctionDef,
)
from guppylang_internals.definition.declaration import RawFunctionDecl
from guppylang_internals.definition.extern import RawExternDef
from guppylang_internals.definition.function import (
    RawFunctionDef,
)
from guppylang_internals.definition.metadata import GuppyMetadata
from guppylang_internals.definition.overloaded import OverloadedFunctionDef
from guppylang_internals.definition.parameter import (
    ConstVarDef,
    RawConstVarDef,
    TypeVarDef,
)
from guppylang_internals.definition.pytket_circuits import (
    RawLoadPytketDef,
    RawPytketDef,
)
from guppylang_internals.definition.struct import RawStructDef
from guppylang_internals.definition.traced import RawTracedFunctionDef
from guppylang_internals.definition.ty import TypeDef
from guppylang_internals.dummy_decorator import _DummyGuppy, sphinx_running
from guppylang_internals.engine import DEF_STORE
from guppylang_internals.span import Loc, SourceMap, Span
from guppylang_internals.tracing.util import hide_trace
from guppylang_internals.tys.arg import Argument
from guppylang_internals.tys.param import Parameter
from guppylang_internals.tys.subst import Inst
from guppylang_internals.tys.ty import (
    FunctionType,
    NoneType,
    NumericType,
    UnitaryFlags,
)
from hugr import ops
from hugr import tys as ht
from hugr import val as hv
from hugr.package import ModulePointer
from typing_extensions import Unpack, dataclass_transform, deprecated

from guppylang.defs import (
    GuppyDefinition,
    GuppyFunctionDefinition,
    GuppyTypeVarDefinition,
)

S = TypeVar("S")
T = TypeVar("T")
F = TypeVar("F", bound=Callable[..., Any])
P = ParamSpec("P")
Decorator = Callable[[S], T]

AnyRawFunctionDef = (
    RawFunctionDef,
    RawCustomFunctionDef,
    RawFunctionDecl,
    RawPytketDef,
    RawLoadPytketDef,
    OverloadedFunctionDef,
)

__all__ = ("guppy", "custom_guppy_decorator", "GuppyKwargs")


class GuppyKwargs(TypedDict, total=False):
    """Typed dictionary specifying the optional keyword arguments for the `@guppy`
    decorator.
    """

    unitary: bool
    control: bool
    dagger: bool
    power: bool
    max_qubits: int


class _Guppy:
    """Class for the `@guppy` decorator."""

    @overload
    def __call__(
        self, /, **kwargs: Unpack[GuppyKwargs]
    ) -> Decorator[Callable[P, T], GuppyFunctionDefinition[P, T]]: ...

    @overload
    def __call__(self, f: Callable[P, T], /) -> GuppyFunctionDefinition[P, T]: ...

    def __call__(
        self, *args: Any, **kwargs: Unpack[GuppyKwargs]
    ) -> (
        GuppyFunctionDefinition[P, T]
        | Decorator[Callable[P, T], GuppyFunctionDefinition[P, T]]
    ):
        def dec(
            f: Callable[P, T], kwargs: GuppyKwargs
        ) -> GuppyFunctionDefinition[P, T]:
            flags, metadata = _parse_kwargs(kwargs)
            defn = RawFunctionDef(
                DefId.fresh(),
                f.__name__,
                None,
                f,
                unitary_flags=flags,
                metadata=metadata,
            )
            DEF_STORE.register_def(defn, get_calling_frame())
            return GuppyFunctionDefinition(defn)

        return _with_optional_kwargs(dec, args, kwargs)

    @overload
    def comptime(
        self, /, **kwargs: Unpack[GuppyKwargs]
    ) -> Decorator[Callable[P, T], GuppyFunctionDefinition[P, T]]: ...

    @overload
    def comptime(self, f: Callable[P, T], /) -> GuppyFunctionDefinition[P, T]: ...

    def comptime(
        self, *args: Any, **kwargs: Unpack[GuppyKwargs]
    ) -> (
        GuppyFunctionDefinition[P, T]
        | Decorator[Callable[P, T], GuppyFunctionDefinition[P, T]]
    ):
        """Registers a function to be executed at compile-time during Guppy compilation,
        enabling the use of arbitrary Python features as long as they don't depend on
        runtime values.

        .. code-block:: python
            from guppylang import guppy
            from guppylang.std.builtins import array

            @guppy.comptime
            def print_arrays(arr1: array[str, 10], arr2: array[str, 10]) -> None:
                for s1, s2 in zip(arr1, arr2):
                    print(f"({s1}, {s2})")
        """

        def dec(
            f: Callable[P, T], kwargs: GuppyKwargs
        ) -> GuppyFunctionDefinition[P, T]:
            _ = _parse_kwargs(kwargs)  # TODO: Pass flags to RawTracedFunctionDef
            defn = RawTracedFunctionDef(DefId.fresh(), f.__name__, None, f)
            DEF_STORE.register_def(defn, get_calling_frame())
            return GuppyFunctionDefinition(defn)

        return _with_optional_kwargs(dec, args, kwargs)

    @deprecated("Use @guppylang_internal.decorator.extend_type instead.")
    def extend_type(self, defn: TypeDef) -> Callable[[type], type]:
        # Set `return_class=True` to match the old behaviour until this deprecated
        # method is removed
        import guppylang_internals.decorator

        return guppylang_internals.decorator.extend_type(defn, return_class=True)

    @deprecated("Use @guppylang_internal.decorator.custom_type instead.")
    def type(
        self,
        hugr_ty: ht.Type | Callable[[Sequence[Argument], CompilerContext], ht.Type],
        name: str = "",
        copyable: bool = True,
        droppable: bool = True,
        bound: ht.TypeBound | None = None,
        params: Sequence[Parameter] | None = None,
    ) -> Callable[[type[T]], type[T]]:
        return custom_type(hugr_ty, name, copyable, droppable, bound, params)

    @dataclass_transform()
    def struct(self, cls: builtins.type[T]) -> builtins.type[T]:
        """Registers a class as a Guppy struct.

        .. code-block:: python
            from guppylang import guppy

            @guppy.struct
            class MyStruct:
                field1: int
                field2: int

            @guppy
            def add_fields(self: "MyStruct") -> int:
                return self.field2 + self.field2
        """
        defn = RawStructDef(DefId.fresh(), cls.__name__, None, cls)
        frame = get_calling_frame()
        DEF_STORE.register_def(defn, frame)
        for val in cls.__dict__.values():
            if isinstance(val, GuppyDefinition):
                DEF_STORE.register_impl(defn.id, val.wrapped.name, val.id)
        # Prior to Python 3.13, the `__firstlineno__` attribute on classes is not set.
        # However, we need this information to precisely look up the source for the
        # class later. If it's not there, we can set it from the calling frame:
        if not hasattr(cls, "__firstlineno__"):
            cls.__firstlineno__ = frame.f_lineno  # type: ignore[attr-defined]
        # We're pretending to return the class unchanged, but in fact we return
        # a `GuppyDefinition` that handles the comptime logic
        return GuppyDefinition(defn)  # type: ignore[return-value]

    def type_var(
        self,
        name: str,
        copyable: bool = True,
        droppable: bool = True,
    ) -> TypeVar:
        """Creates a new type variable.

        .. code-block:: python
            from guppylang import guppy

            T = guppy.type_var("T")

            @guppy
            def identity(x: T) -> T:
                return x
        """
        defn = TypeVarDef(DefId.fresh(), name, None, copyable, droppable)
        DEF_STORE.register_def(defn, get_calling_frame())
        # We're pretending to return a `typing.TypeVar`, but in fact we return a special
        # `GuppyDefinition` that pretends to be a TypeVar at runtime
        return GuppyTypeVarDefinition(defn, TypeVar(name))  # type: ignore[return-value]

    def nat_var(self, name: str) -> TypeVar:
        """Creates a new nat variable."""
        defn = ConstVarDef(DefId.fresh(), name, None, NumericType(NumericType.Kind.Nat))
        DEF_STORE.register_def(defn, get_calling_frame())
        # We're pretending to return a `typing.TypeVar`, but in fact we return a special
        # `GuppyDefinition` that pretends to be a TypeVar at runtime
        return GuppyTypeVarDefinition(defn, TypeVar(name))  # type: ignore[return-value]

    def const_var(self, name: str, ty: str) -> TypeVar:
        """Creates a new const type variable."""
        type_ast = _parse_expr_string(
            ty, f"Not a valid Guppy type: `{ty}`", DEF_STORE.sources
        )
        defn = RawConstVarDef(DefId.fresh(), name, None, type_ast)
        DEF_STORE.register_def(defn, get_calling_frame())
        # We're pretending to return a `typing.TypeVar`, but in fact we return a special
        # `GuppyDefinition` that pretends to be a TypeVar at runtime
        return GuppyTypeVarDefinition(defn, TypeVar(name))  # type: ignore[return-value]

    @deprecated("Use @guppylang_internal.decorator.custom_function instead.")
    def custom(
        self,
        compiler: CustomInoutCallCompiler | None = None,
        checker: CustomCallChecker | None = None,
        higher_order_value: bool = True,
        name: str = "",
        signature: FunctionType | None = None,
    ) -> Callable[[Callable[P, T]], GuppyFunctionDefinition[P, T]]:
        return custom_function(compiler, checker, higher_order_value, name, signature)

    @deprecated("Use @guppylang_internal.decorator.hugr_op instead.")
    def hugr_op(
        self,
        op: Callable[[ht.FunctionType, Inst, CompilerContext], ops.DataflowOp],
        checker: CustomCallChecker | None = None,
        higher_order_value: bool = True,
        name: str = "",
        signature: FunctionType | None = None,
    ) -> Callable[[Callable[P, T]], GuppyFunctionDefinition[P, T]]:
        return hugr_op(op, checker, higher_order_value, name, signature)

    @overload
    def declare(
        self, /, **kwargs: Unpack[GuppyKwargs]
    ) -> Decorator[Callable[P, T], GuppyFunctionDefinition[P, T]]: ...

    @overload
    def declare(self, f: Callable[P, T], /) -> GuppyFunctionDefinition[P, T]: ...

    def declare(
        self, *args: Any, **kwargs: Unpack[GuppyKwargs]
    ) -> (
        GuppyFunctionDefinition[P, T]
        | Decorator[Callable[P, T], GuppyFunctionDefinition[P, T]]
    ):
        """Declares a Guppy function without defining it."""

        def dec(
            f: Callable[P, T], kwargs: GuppyKwargs
        ) -> GuppyFunctionDefinition[P, T]:
            flags, _ = _parse_kwargs(kwargs)
            defn = RawFunctionDecl(
                DefId.fresh(), f.__name__, None, f, unitary_flags=flags
            )
            DEF_STORE.register_def(defn, get_calling_frame())
            return GuppyFunctionDefinition(defn)

        return _with_optional_kwargs(dec, args, kwargs)

    def overload(
        self, *funcs: Any
    ) -> Callable[[Callable[P, T]], GuppyFunctionDefinition[P, T]]:
        """Collects multiple function definitions into one overloaded function.

        Consider the following example:

        .. code-block:: python

            @guppy.declare
            def variant1(x: int, y: int) -> int: ...

            @guppy.declare
            def variant2(x: float) -> int: ...

            @guppy.overload(variant1, variant2)
            def combined(): ...


        Now, `combined` may be called with either one `float` or two `int` arguments,
        delegating to the implementation with the matching signature:

        .. code-block:: python

            combined(4.2)  # Calls `variant1`
            combined(42, 43)  # Calls `variant2`


        Note that the compiler will pick the *first* implementation with matching
        signature and ignore all following ones, even if they would also match. For
        example, if we added a third variant

        .. code-block:: python

            @guppy.declare
            def variant3(x: int) -> int: ...

            @guppy.overload(variant1, variant2, variant3)
            def combined_new(): ...

        then a call `combined_new(42)` will still select the `variant1` implementation
        `42` is a valid argument for `variant1` and `variant1` comes before `variant3`
        in the `@guppy.overload` annotation.
        """
        funcs = list(funcs)
        if len(funcs) < 2:
            raise ValueError("Overload requires at least two functions")
        func_ids = []
        for func in funcs:
            if not isinstance(func, GuppyDefinition):
                raise TypeError(f"Not a Guppy definition: {func}")
            if not isinstance(func.wrapped, AnyRawFunctionDef):
                raise TypeError(
                    f"Not a Guppy function definition: {func.wrapped.description} "
                    f"`{func.wrapped.name}`"
                )
            func_ids.append(func.id)

        def dec(f: Callable[P, T]) -> GuppyFunctionDefinition[P, T]:
            dummy_sig = FunctionType([], NoneType())
            defn = OverloadedFunctionDef(
                DefId.fresh(), f.__name__, None, dummy_sig, func_ids
            )
            DEF_STORE.register_def(defn, get_calling_frame())
            return GuppyFunctionDefinition(defn)

        return dec

    def constant(self, name: str, ty: str, value: hv.Value) -> T:  # type: ignore[type-var]  # Since we're returning a free type variable
        """Adds a constant to a module, backed by a `hugr.val.Value`."""
        type_ast = _parse_expr_string(
            ty, f"Not a valid Guppy type: `{ty}`", DEF_STORE.sources
        )
        defn = RawConstDef(DefId.fresh(), name, None, type_ast, value)
        DEF_STORE.register_def(defn, get_calling_frame())
        # We're pretending to return a free type variable, but in fact we return
        # a `GuppyDefinition` that handles the comptime logic
        return GuppyDefinition(defn)  # type: ignore[return-value]

    def _extern(
        self,
        name: str,
        ty: str,
        symbol: str | None = None,
        constant: bool = True,
    ) -> T:  # type: ignore[type-var]  # Since we're returning a free type variable
        """Adds an extern symbol to a module."""
        type_ast = _parse_expr_string(
            ty, f"Not a valid Guppy type: `{ty}`", DEF_STORE.sources
        )
        defn = RawExternDef(
            DefId.fresh(), name, None, symbol or name, constant, type_ast
        )
        DEF_STORE.register_def(defn, get_calling_frame())
        # We're pretending to return a free type variable, but in fact we return
        # a `GuppyDefinition` that handles the comptime logic
        return GuppyDefinition(defn)  # type: ignore[return-value]

    @deprecated(
        "guppy.compile(foo) is deprecated and will be removed in a future version:"
        " use foo.compile() instead."
    )
    def compile(self, obj: Any) -> ModulePointer:
        """Compiles a Guppy definition to Hugr."""

        if not isinstance(obj, GuppyDefinition):
            raise TypeError(f"Object is not a Guppy definition: {obj}")
        return ModulePointer(obj.compile(), 0)

    def pytket(
        self, input_circuit: Any
    ) -> Callable[[Callable[P, T]], GuppyFunctionDefinition[P, T]]:
        """Backs a function declaration by the given pytket circuit. The declaration
        signature needs to match the circuit definition in terms of number of qubit
        inputs and measurement outputs.

        There is no linearity checking inside pytket circuit functions. Any measurements
        inside the circuit get returned as bools, but the qubits do not get consumed and
        the  pytket circuit function does not require ownership. You should either make
        sure you discard all qubits you know are measured during the circuit, or avoid
        measurements in the circuit and measure in Guppy afterwards.

        Note this decorator doesn't support passing inputs as arrays (use `load_pytket`
        instead).

        .. code-block:: python
            from pytket import Circuit
            from guppylang import guppy

            circ = Circuit(1)
            circ.H(0)
            circ.measure_all()

            @guppy.pytket(circ)
            def guppy_circ(q: qubit) -> bool: ...

            @guppy
            def foo(q: qubit) -> bool:
                return guppy_circ(q)"""

        err_msg = "Only pytket circuits can be passed to guppy.pytket"
        try:
            import pytket

            if not isinstance(input_circuit, pytket.circuit.Circuit):
                raise TypeError(err_msg) from None

        except ImportError:
            raise TypeError(err_msg) from None

        def func(f: Callable[P, T]) -> GuppyFunctionDefinition[P, T]:
            defn = RawPytketDef(DefId.fresh(), f.__name__, None, f, input_circuit)
            DEF_STORE.register_def(defn, get_calling_frame())
            return GuppyFunctionDefinition(defn)

        return func

    def load_pytket(
        self,
        name: str,
        input_circuit: Any,
        *,
        use_arrays: bool = True,
    ) -> GuppyFunctionDefinition[..., Any]:
        """Load a pytket :py:class:`~pytket.circuit.Circuit` as a Guppy function. By
        default, each qubit register is represented by an array input (and each bit
        register as an array output), with the order being determined lexicographically.
        The default registers are 'q' and 'c' respectively. You can disable array usage
        and pass individual qubits by passing `use_arrays=False`.

        .. code-block:: python

            from pytket import Circuit
            from guppylang import guppy

            circ = Circuit(2)
            reg = circ.add_q_register("extra_reg", 3)
            circ.measure_register(reg, "extra_bits")

            guppy_circ = guppy.load_pytket("guppy_circ", circ)

            @guppy
            def foo(default_reg: array[qubit, 2],
                    extra_reg: array[qubit, 3]) -> array[bool, 3]:
                # Note that the default_reg name is 'q' so it has to come after 'e...'
                # lexicographically.
                return guppy_circ(extra_reg, default_reg)

        Any symbolic parameters in the circuit need to be passed as a lexicographically
        sorted array (if arrays are enabled, else individually in that order) as values
        of type `angle`.

        The function name is determined by the function variable you bind the `
        load_pytket`method call to, however the name string passed to the method should
        match this variable for error reporting purposes.

        There is no linearity checking inside pytket circuit functions. Any measurements
        inside the circuit get returned as bools, but the qubits do not get consumed and
        the  pytket circuit function does not require ownership. You should either make
        sure you discard all qubits you know are measured during the circuit, or avoid
        measurements in the circuit and measure in Guppy afterwards.
        """

        err_msg = "Only pytket circuits can be passed to guppy.load_pytket"
        try:
            import pytket

            if not isinstance(input_circuit, pytket.circuit.Circuit):
                raise TypeError(err_msg) from None

        except ImportError:
            raise TypeError(err_msg) from None

        span = _find_load_call(DEF_STORE.sources)
        defn = RawLoadPytketDef(
            DefId.fresh(), name, None, span, input_circuit, use_arrays
        )
        DEF_STORE.register_def(defn, get_calling_frame())
        return GuppyFunctionDefinition(defn)


def _parse_expr_string(ty_str: str, parse_err: str, sources: SourceMap) -> ast.expr:
    """Helper function to parse expressions that are provided as strings.

    Tries to infer the source location were the given string was defined by inspecting
    the call stack.
    """
    try:
        expr_ast = ast.parse(ty_str, mode="eval").body
    except SyntaxError:
        raise SyntaxError(parse_err) from None

    # Try to annotate the type AST with source information. This requires us to
    # inspect the stack frame of the caller
    if caller_frame := get_calling_frame():
        info = inspect.getframeinfo(caller_frame)
        if caller_module := inspect.getmodule(caller_frame):
            sources.add_file(info.filename)
            source_lines, _ = inspect.getsourcelines(caller_module)
            source = "".join(source_lines)
            annotate_location(expr_ast, source, info.filename, 1)
            # Modify the AST so that all sub-nodes span the entire line. We
            # can't give a better location since we don't know the column
            # offset of the `ty` argument
            for node in [expr_ast, *ast.walk(expr_ast)]:
                node.lineno = node.end_lineno = info.lineno
                node.col_offset = 0
                node.end_col_offset = len(source_lines[info.lineno - 1]) - 1
    return expr_ast


def _find_load_call(sources: SourceMap) -> Span | None:
    """Helper function to find location where pytket circuit was loaded.

    Tries to define a source code span by inspecting the call stack.
    """
    # Go back as first frame outside of compiler modules is 'pretty_errors_wrapped'.
    if load_frame := get_calling_frame():
        info = inspect.getframeinfo(load_frame)
        filename = info.filename
        lineno = info.lineno
        sources.add_file(filename)
        # If we don't support python <= 3.10, this can be done better with
        # info.positions which gives you exact offsets.
        # For now over approximate and make the span cover the entire line.
        if load_module := inspect.getmodule(load_frame):
            source_lines, _ = inspect.getsourcelines(load_module)
            max_offset = len(source_lines[lineno - 1]) - 1

            start = Loc(filename, lineno, 0)
            end = Loc(filename, lineno, max_offset)
            return Span(start, end)
    return None


def custom_guppy_decorator(f: F) -> F:
    """Decorator to mark user-defined decorators that wrap builtin `guppy` decorators.

    Example:

    .. code-block:: python

        @custom_guppy_decorator
        def my_guppy(f):
            # Some custom logic here ...
            return guppy(f)

        @my_guppy
        def main() -> int: ...

    If the `custom_guppy_decorator` were missing, then the `@my_guppy` annotation would
    not produce a valid guppy definition.
    """
    f.__code__ = f.__code__.replace(co_name="__custom_guppy_decorator__")
    return f


def get_calling_frame() -> FrameType:
    """Finds the first frame that called this function outside the compiler modules."""
    frame = inspect.currentframe()
    while frame:
        # Skip frame if we're inside a user-defined decorator that wraps the `guppy`
        # decorator. Those are functions with a special `__code__.co_name` of
        # "__custom_guppy_decorator__".
        if frame.f_code.co_name == "__custom_guppy_decorator__":
            frame = frame.f_back
            continue
        module = inspect.getmodule(frame)
        if module is None:
            return frame
        if module.__file__ != __file__:
            return frame
        frame = frame.f_back
    raise RuntimeError("Couldn't obtain stack frame for definition")


def _with_optional_kwargs(
    decorator: Callable[[S, GuppyKwargs], T], args: tuple[Any, ...], kwargs: GuppyKwargs
) -> T | Callable[[S], T]:
    """Helper function to define decorators that may be used directly (`@decorator`) but
    also with optional keyword arguments (`@decorator(kwarg=value)`).
    """
    match args:
        case ():
            return lambda f: decorator(f, kwargs)
        case (f,):
            if kwargs:
                err = "Unexpected keyword arguments"
                raise TypeError(err)
            return decorator(f, kwargs)
        case _:
            err = "Unexpected positional arguments"
            raise TypeError(err)


@hide_trace
def _parse_kwargs(kwargs: GuppyKwargs) -> tuple[UnitaryFlags, GuppyMetadata]:
    """Parses the kwargs dict specified in the `@guppy` decorator into `UnitaryFlags`
    and other metadata that will be passed onto the compiled function as is.
    """
    flags = UnitaryFlags.NoFlags
    if kwargs.pop("unitary", False):
        flags |= UnitaryFlags.Unitary
    if kwargs.pop("control", False):
        flags |= UnitaryFlags.Control
    if kwargs.pop("dagger", False):
        flags |= UnitaryFlags.Dagger
    if kwargs.pop("power", False):
        flags |= UnitaryFlags.Power

    metadata = GuppyMetadata()
    metadata.max_qubits.value = kwargs.pop("max_qubits", None)

    if remaining := next(iter(kwargs), None):
        err = f"Unknown keyword argument: `{remaining}`"
        raise TypeError(err)
    return flags, metadata


guppy = cast(_Guppy, _DummyGuppy()) if sphinx_running() else _Guppy()
